package ach;

import java.awt.*;
import java.awt.image.*;
import java.applet.Applet;
import java.lang.Math;
import java.io.File;
import java.io.FileWriter;
import java.net.URL;
import java.util.Hashtable;
import java.util.StringTokenizer;
import ach.CatalogPanel;            // class to handle thumbnail catalogs
import ach.DemoLocal;               // class containing language dependent string constants
import ach.SmartScrollPane;         // a class similar to Java 1.1 ScrollPane (works with Java 1.0)
import ach.ImgCanvas;               // an extended Canvas
import ach.ImgComponent;            // alternatively used instead of ImgCanvas as lightweight component
import ach.IconButton;              // Image button
import ach.MinLabel;                // Component derived from Label with optimized sizing
import ach.ToolbarController;       // interface for toolbar controls
import ach.image.EMemImageSource;   // extended MemoryImageSource
import ach.image.ImageInfo;         // additional image related information
import ach.image.PreviewOpenDialog;
import ach.image.PrintImage;        // class to control printing
import ach.image.PrintImage2;       // class to control printing with Java 2 (JDK 1.2 and higher)
import ach.image.PrintDialog;       // provides a printer dialog
import ach.image.ImgProdObserver;   // an observer for extended imageproducers
import ach.image.BatchPrinter;      // for callback communication to class PrintImage
import ach.file.ParamUtil;          // some parameter and path oriented utilities
import ach.types.PRectangle;        // Rectangle with 2D-point coordinates for upper left and lower right corner
import ach.vectorGraphics.*;        // classes with vectorgraphic objects to draw annotations

/**
 * ImagePanel.java                                                          <br>
 * Copyright (C) 1997-2007 H. Hackbarth, All Rights Reserved                <br>
 * Panel used as container to display images. Contains scrollbars and several
 * display control buttons.
 * For examples how to use this class please refer to TiffyApplet.java and
 * SwingTiffy.java. Parameter options can be found in method parseOptions().
 * ImagePanel embeds either the heavyweight class ImgCanvas or the
 * lightweight class ImgComponent to handle the image display. The code
 * contains remarks (search for "lightweight") how to do the required
 * replacements.                                                            <br>
 *                                                                          <br>
 * THE AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY
 * OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. THE AUTHOR SHALL NOT BE LIABLE FOR ANY
 * DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 *
 * @author Helge Hackbarth
 * @version 3.1.5
 */
public class ImagePanel extends Panel implements ImgProdObserver, BatchPrinter, ToolbarController, Scrollable, AnnotationSaver
// public class ImagePanel extends javax.swing.JPanel implements ImgProdObserver, BatchPrinter, ToolbarController, Scrollable, AnnotationSaver // lightweight replacement
{
  // general variables for demo panel instance
  private Container parent;       // the parent container that was passed via constructor
  private boolean versionEqualGreater1_1 = // flag used for version dependent behaviour
    (System.getProperty("java.version").compareTo("1.1") >= 0);
  private boolean scaledInstanceSupported = // may be set false later by exception
    versionEqualGreater1_1;     // as not all 1.1 VMs support getScaledInstance
  private String errorMsg;            // keeps errorMsg of last operation

  // variables to handle the image data and related information
  EMemImageSource imgSrc = null;       // extended MemoryImageSource for current image
  ImageInfo imgInfo = new ImageInfo(); // to keep additional image information

  // SmartScrollPane (provides functionality similar to ScrollPane already for Java 1.0)
  SmartScrollPane sScrPane;

  // Display/printing control parameters
  private double xZoom = 1;                  // horizontal zoom factor
  private double yZoom = 1;                  // vertical zoom factor
  private boolean fitHoriz = false;          // true if image is to be zoomed to horizontal window width
  private boolean fitInWindow = false;       // true if image is to be zoomed proportionally to window
  private boolean cmdLinePrint = false;      // true if image is to be printed automatically
  private boolean quickPrint = false;        // true if image is to be printed immediately without Tiffy's print dialog
  private boolean saveEnabled = false;       // flag if save button should be enabled (default: disabled)
  private boolean paletteCorrected = false;  // flag if palette was corrected (needed for Java 1.0)
  private int scaleHint = Image.SCALE_FAST;  // type of scaling to be used (fast or smooth)
  private  Image tmpImg = null;               // store visible area as temporary image
  private Rectangle tmpImgRect = null;       // bounding rectangle of visible area after previous operation
  private Rectangle newTmpImgRect;           // bounding rectangle of new visible area
  /**
   * flag needed due to bug in MS VM version 2334; see example SwingTiffy.java
   */
  public boolean frameUpdated = false;       // flag needed due to bug in MS VM version 2334
  private boolean doInvert = false;          // true if parameter -i for inversion was passed
  private int doRotate = 0;                  // status of parameter -r: 0=no rot.; 1=90; 2=180; 3=270
  private boolean autorotateAllowed = true;  // false if parameter -noautorotate was passed
  private boolean unskewAllowed = true;      // flag if skewed images should be scaled to square pixels
  private boolean annotationsEditable = true;// flag if annotationpanel is allowed to appear
  private boolean showAnnotations = true;    // flag if annotations should be shown if found
  private String bufferedAnnotations;        // annotions may be buffered for scheduled processing
  private boolean showCatalog = true;        // flag if thumbnail catalog should be displayed
  private CatalogPanel catPanel = null;      // panel containing thumbnail catalog (if enabled)
  private String[] catSlides = null;         // will be filled with slides from filename list
  private int thumbNailWidth = 192;          // specifies dimension of thumbnails in catalog panel
  private int thumbNailHeight = 256;         // specifies dimension of thumbnails in catalog panel
  final static int CATALOGFINISHED = 99993;  // user defined event ID for finished catalog
  private NavDialog navDialog = null;        // Navigation dialog
  private Frame dummyFrame = new Frame("");  // temporary parent for dialog boxes
  private Hashtable props = new Hashtable(); // keeps display/print properties
  private Object printJob = null;            // globally accessible PrintJob or PrinterJob (can be used so as of Java 2)
  private Object book = null;                // Book for multipage printing with Java 2
  private int printPageNum;                  // current page to be printed
  private String printPageSelection = "";    // specification of pages to be printed
  private int preferredResolution = 0;       // preferred print resolution (0 = default)

  // variables to control slide sequence
  private String[] slides = null;            // can keep a slide sequence
  private int slideIdx = -1;                 // index of current slide

  // use imgCanvas for extended image display functionality
  private ImgComponentIntf imgCanvas = null;
  private Panel imgPanel = new Panel();      // panel where the imgCanvas will be embedded
  // private javax.swing.JPanel imgPanel = new javax.swing.JPanel(); // lightweight JPanel where the imgCanvas will be embedded

  // Mouse action control parameters
  //--------------------------------------------------------------------------
  private boolean mouseDown = false;
  private Point mouseStart = new Point(0,0); // coordinates where mouse dragging startet
  private Point mouseMove = new Point(0,0);  // last mouse location

  // Define button panel
  //--------------------------------------------------------------------------
  Panel buttonPanel         = new Panel();
  IconButton btnSave        = new IconButton("save.gif",DemoLocal.s_Save + " (" + DemoLocal.s_Download + ")");
  IconButton btnZoomIn      = new IconButton("zoomplus.gif",DemoLocal.s_zoomIn);
  IconButton btnZoomOut     = new IconButton("zoominus.gif",DemoLocal.s_zoomOut);
  IconButton btnNavigation  = new IconButton("grabhand.gif",DemoLocal.s_navigation);
  TextField tfZoom          = new TextField("100%", 3); // columns = 3 results in a width of about 5 chars
  IconButton btn1to1        = new IconButton("1to1.gif",DemoLocal.s_origSize);
  IconButton btnFitHoriz    = new IconButton("horscale.gif",DemoLocal.s_fitHorizWindow);
  IconButton btnFitInWindow = new IconButton("winscale.gif",DemoLocal.s_fitInWindow);
  IconButton btnPrevPage    = new IconButton("pgback.gif",DemoLocal.s_pageBack);
  TextField tfPageNum       = new TextField("1", 1); // columns = 1 results in a width of about 4 chars
  MinLabel lNumPages        = new MinLabel();
  IconButton btnNextPage    = new IconButton("pgfwd.gif",DemoLocal.s_pageFwd);
  IconButton btnThumbnails  = new IconButton("thumbnails.gif",DemoLocal.s_preview);
  IconButton btnRot090      = new IconButton("turnrght.gif",DemoLocal.s_rot090);
  IconButton btnRot180      = new IconButton("turn180.gif",DemoLocal.s_rot180);
  IconButton btnRot270      = new IconButton("turnleft.gif",DemoLocal.s_rot270);
  IconButton btnMirror      = new IconButton("mirror.gif",DemoLocal.s_mirror);
  IconButton btnFlip        = new IconButton("flip.gif",DemoLocal.s_flip);
  IconButton btnScaleSmooth = new IconButton("antialia.gif",DemoLocal.s_scaleSmooth);
  IconButton btnInvert      = new IconButton("invert.gif",DemoLocal.s_invert);
  IconButton btnAnnotations = new IconButton("annotation.gif",DemoLocal.s_annotations);
  IconButton btnPrint       = new IconButton("printer.gif",DemoLocal.s_print);
  IconButton btnPrintView   = new IconButton("printview.gif",DemoLocal.s_printView);
  IconButton btnHelp        = new IconButton("help.gif",DemoLocal.s_help);
  String buttonPanelCtrl; // controls through options which button is visible/hidden

  // annotation panel and other annotation related vars
  //--------------------------------------------------------------------------
  Panel annotationPanel     = new Panel();
  ButtonGroup annotationButtonGroup = new ButtonGroup();
  IconButton btnPointer     = new IconButton("pointer.gif",DemoLocal.s_Annotations + " " + DemoLocal.s_select);
  IconButton btnRectangle   = new IconButton("filledrect.gif",DemoLocal.s_Rectangle);
  IconButton btnOval        = new IconButton("filledoval.gif",DemoLocal.s_Oval);
  IconButton btnPolygon     = new IconButton("freehand.gif",DemoLocal.s_Polygon + "/" + DemoLocal.s_Freehand);
  IconButton btnText        = new IconButton("text.gif",DemoLocal.s_Text);
  IconButton btnMemo        = new IconButton("memo.gif",DemoLocal.s_Memo);
  IconButton btnStamp       = new IconButton("stamp.gif",DemoLocal.s_Stamp);
  IconButton btnShapeProps  = new IconButton("config.gif",DemoLocal.s_Settings);
  IconButton btnSaveAnnot   = new IconButton("save.gif",DemoLocal.s_Save + " " + DemoLocal.s_Annotations);
  Label toolHelpLbl         = new Label(DemoLocal.s_PointerHelp);
  GraphicEditor graphEd     = new GraphicEditor(dummyFrame);
  private AnnotationSaver annotationSaver = this; // Points to currently used annotationSaver implementation

  // Define "component pointers" used to control behavior
  // of buttons for previous and next page
  //--------------------------------------------------------------------------
  Component cSave     = btnSave; // let it point to save button
  Component cPrint    = btnPrint; // let it point to print button
  Component cPrintView = btnPrintView; // let it point to print view button
  Component cPrevPage = btnPrevPage; // let it point to previous page button
  Component cNextPage = btnNextPage; // let it point to next page button
  Component cThumbnails = btnThumbnails; // let it point to thumbnails button
  Component cAnnotations = btnAnnotations; // let it point to annotations button
  Component cScaleSmooth = btnScaleSmooth; // let it point to ScaleSmooth button
  ToolbarController tbc = this;
  Dimension initialSize = null;
  
  GcTimer gcTimer = null; // garbage collection timer


  void addButtonPanel() {
    buttonPanel.setBackground(Color.lightGray);
    // set up button panel using GridBagLayout
    GridBagLayout gb = new GridBagLayout();
    GridBagConstraints gc = new GridBagConstraints();
    buttonPanel.setLayout(gb);
    gc.insets = new Insets(2,2,2,2); // leave space for 3D-border
    gc.anchor = GridBagConstraints.WEST;
    buttonPanel.add(btnSave);
    if (!saveEnabled)
      btnSave.disable();
    buttonPanel.add(btnPrint);
    buttonPanel.add(btnPrintView);
    buttonPanel.add(new IconButton(null)); // seperator
    buttonPanel.add(btnNavigation);
    buttonPanel.add(btnZoomIn);
    buttonPanel.add(btnZoomOut);
    buttonPanel.add(tfZoom);
    buttonPanel.add(btn1to1);
    buttonPanel.add(btnFitHoriz);
    buttonPanel.add(btnFitInWindow);
    buttonPanel.add(new IconButton(null)); // seperator
    buttonPanel.add(btnPrevPage);
    buttonPanel.add(tfPageNum);
    buttonPanel.add(lNumPages);
    buttonPanel.add(btnNextPage);
    buttonPanel.add(btnThumbnails);
    btnThumbnails.setToggleMode(true); // configure this button as two state button
    buttonPanel.add(new IconButton(null)); // seperator
    buttonPanel.add(btnRot090);
    buttonPanel.add(btnRot180);
    buttonPanel.add(btnRot270);
    buttonPanel.add(btnMirror);
    buttonPanel.add(btnFlip);
    buttonPanel.add(new IconButton(null)); // seperator
    buttonPanel.add(btnScaleSmooth);
    btnScaleSmooth.setToggleMode(true); // configure this button as two state button
    buttonPanel.add(btnInvert);
    buttonPanel.add(btnAnnotations);
    btnAnnotations.setToggleMode(true); // configure this button as two state button
    btnAnnotations.enable(); // initially disable
    buttonPanel.add(new IconButton(null)); // seperator
    buttonPanel.add(btnHelp);
    // Font newFnt = new Font("Helvetica", Font.PLAIN, 12);
    for (int i = 0; i < buttonPanel.countComponents(); i++) {
      Component c = buttonPanel.getComponent(i);
    //   c.setFont(newFnt);
    //   gb.setConstraints(c,gc);
      if ((buttonPanelCtrl != null) && (buttonPanelCtrl.length()>i))
        c.setVisible(!buttonPanelCtrl.substring(i,i+1).equals("0"));
    }
    // embed button panel into another panel to left-adjust the buttons
    Panel p = new Panel();
    p.setLayout(new BorderLayout());
    p.add("West", buttonPanel);
    add("North", p);
  }

  void addAnnotationPanel() {
    annotationPanel.setBackground(Color.lightGray);
    // set up button panel using GridBagLayout
    GridBagLayout gb = new GridBagLayout();
    GridBagConstraints gc = new GridBagConstraints();
    annotationPanel.setLayout(gb);
    gc.insets = new Insets(2,2,2,2); // leave space for 3D-border
    gc.anchor = GridBagConstraints.WEST;
    annotationPanel.add(btnPointer);
    annotationPanel.add(btnRectangle);
    annotationPanel.add(btnOval);
    annotationPanel.add(btnPolygon);
    annotationPanel.add(btnText);
    annotationPanel.add(btnMemo);
    annotationPanel.add(btnStamp);
    annotationPanel.add(btnShapeProps);
    annotationPanel.add(btnSaveAnnot);
    for (int i=0; i<annotationPanel.countComponents(); i++) {
      IconButton ib = (IconButton)annotationPanel.getComponent(i);
      ib.enable();
      if (ib != btnSaveAnnot)
        ib.setToggleMode(true); // configure this button as two state button
      annotationButtonGroup.add(ib);
    }
    btnPointer.setToggleState(true);
    setCanvasCursor(Cursor.DEFAULT_CURSOR);
    graphEd.setMode(GraphicEditor.POINTERMODE, null);
    graphEd.setStamps((Stamp[])props.get("STAMPS"));
    if (graphEd.getStamps() == null) {
      Stamp[] stamps = new Stamp[2];
      stamps[0] = new Stamp("Approved", Stamp.IMAGE, "images/approved.gif");
      stamps[1] = new Stamp("DateTime",Stamp.TEXT,"%DATE\r\n%TIME");
      graphEd.setStamps(stamps);
    }
    Integer tmpInt = (Integer)props.get("CURRENTSTAMP");
    if (tmpInt != null)
      graphEd.currentStampIdx = tmpInt.intValue();
    ShapeProps tmpShapeProps = (ShapeProps)props.get("SHAPEPROPS");
    if (tmpShapeProps != null)
      graphEd.currentShapeProps = tmpShapeProps;
    // embed button panel into another panel to left-adjust the buttons
    Panel p = new Panel();
    p.setLayout(new BorderLayout());
    p.add("West", annotationPanel);
    p.add("Center", toolHelpLbl);
    add("South", p);
  }

  // scan option string passed from parent for known options
  // and initialize corresponding variables
  //--------------------------------------------------------------------------
  void parseOptions(String options) {
    props = ParamUtil.loadProps(parent); // load user-properties (like annotation settings, print margins etc.)
    String[] parameters = ParamUtil.retrieveParameters(options.toUpperCase());
    for (int i=0; i < parameters.length; i++) {
      String ps = parameters[i];
      if (ps.startsWith("-") || ps.startsWith("/"))
        switch (Character.toUpperCase(ps.charAt(1))) {
          case '1':
            if ((ps.length() >= 7) && (ps.substring(1,7).equalsIgnoreCase("150DPI")))
              props.put("RESOLUTION","150 dpi");
            break;
          case 'I':
            doInvert = true;
            break;
          case 'F': // make image fit in horizontal
            if ((ps.length() >= 3) && (ps.substring(1,3).equalsIgnoreCase("FH")))
              fitHoriz = true;
            else // make image fit in window
              if ((ps.length() >= 3) && (ps.substring(1,3).equalsIgnoreCase("FW")))
                fitInWindow = true;
            break;
          case 'L':
            if ((ps.length() >= 11) && (ps.substring(1,11).equalsIgnoreCase("LIGHTPRINT"))) // rotate
              props.put("FORCEJDK11PRINTJOB", new Boolean(true));
            break;
          case 'P':
            // /page : specify page in a multi document image file
            if ((ps.length() >= 5) && (ps.substring(1,5).equalsIgnoreCase("PAGE"))) {
              String pageStr = ps.substring(5);
              if ((pageStr != null) && (pageStr.length() > 0)) {
                imgInfo.page = Long.valueOf(pageStr).intValue();
                if (imgInfo.page <= 1)
                  imgInfo.page = 1;
              } else
                imgInfo.page = 1;
            }
            // /print : open print dialog automatically after image is loaded
            //          and call cleanUp() after image is printed.
            //          Processed asynchronously in method imageProducerUpdate()
            else if ((ps.length() == 6) && (ps.substring(1,6).equalsIgnoreCase("PRINT")))
              cmdLinePrint = true;
            break;
          case 'Q':
            if ((ps.length() >= 11) && (ps.substring(1,11).equalsIgnoreCase("QUICKPRINT"))) // buttons
              quickPrint = true;
            break;
          case 'R': // rotation
            doRotate = Long.valueOf(ps.substring(2)).intValue();
            break;
          case 'S': // enable saving (downloading) of images
            if ((ps.length() >= 5) && (ps.substring(1,5).equalsIgnoreCase("SAVE")))
              saveEnabled = true;
            else if ((ps.length() >= 7) && (ps.substring(1,7).equalsIgnoreCase("SMOOTH")))
              scaleHint = Image.SCALE_SMOOTH;
            break;
          case 'N': // disable gauge, printing, annotations
            if ((ps.length() >= 8) && (ps.substring(1,8).equalsIgnoreCase("NOGAUGE")))
              Gauge.visible = false;
            else if ((ps.length() >= 8) && (ps.substring(1,8).equalsIgnoreCase("NOPRINT"))) {
              cPrint.disable();
              cPrintView.disable();
            }
            else if ((ps.length() >= 13) && (ps.substring(1,13).equalsIgnoreCase("NOTHUMBNAILS")))
              showCatalog = false;
            else if ((ps.length() >= 13) && (ps.substring(1,13).equalsIgnoreCase("NOAUTOROTATE")))
              autorotateAllowed = false;
            else if ((ps.length() >= 14) && (ps.substring(1,14).equalsIgnoreCase("NOANNOTATIONS")))
              imgInfo.annotationsAllowed = false;
            break;
          case 'B':
            if ((ps.length() >= 8) && (ps.substring(1,8).equalsIgnoreCase("BUTTONS"))) // buttons
              buttonPanelCtrl = ps.substring(8);
            break;
        }
    }
    if (props.get("FORCEJDK11PRINTJOB") == null) // make JDK11-printing default
      props.put("FORCEJDK11PRINTJOB", new Boolean(true));

  }

  // check if filename string contains more than one filename and
  // create string array containing slide sequence
  void parseSlides() {
    if ((imgInfo.imgName != null) && (imgInfo.imgName.indexOf(";") >= 0)) {
      if (imgInfo.imgURL == null)
        slides = ParamUtil.retrieveParameters(imgInfo.imgName, ';');
      else
        slides = ParamUtil.retrieveParameters(imgInfo.imgURL.toString(), ';');
      if (slides != null) {
        catSlides = slides;
        if (imgInfo.page > 1) { // check if optional page specified
          slideIdx = imgInfo.page - 1;
          if (slideIdx >= slides.length)
            slideIdx = 0;
        } else slideIdx = 0;
        imgInfo = slideToImageInfo();
      }
    }
  }

  // retrieve imgName / imgURL and page from slide parameters
  // imgInfo.imgURL == null indicates application running standalone
  ImageInfo slideToImageInfo() {
    ImageInfo newImgInfo = new ImageInfo();
    String slideName = slides[slideIdx];
    newImgInfo.imgFileBA = null;
    if (slideName.endsWith("]")) {
      int li = slideName.lastIndexOf('[');
      if (li > 0) {
        String pageStr = slideName.substring(li + 1, slideName.length()-1);
        try {
          newImgInfo.page = Integer.parseInt(pageStr);
        }  catch (Exception e) { }
        slideName = slideName.substring(0,li);
      }
    }
    boolean convertToURL = false;
    try {
      convertToURL = (((imgInfo.imgURL == null) && (slideName.indexOf("://") > 0)) ||
                      (imgInfo.imgURL != null) ||
                      (((java.applet.Applet)parent).getDocumentBase() != null));
    } catch (Exception e) { } // Exception if not running standalone (not as Applet)
    if (convertToURL) {  // assuming that parent is an instance of class Applet
      try {
        newImgInfo.imgName = slideName;
        newImgInfo.imgURL = new java.net.URL(
          ((java.applet.Applet)parent).getDocumentBase(), slideName);
      } catch (Throwable e) { // parent not instance of class Applet?
        try {
          newImgInfo.imgName = slideName;
          newImgInfo.imgURL = new java.net.URL(slideName);
        } catch (java.net.MalformedURLException me) { }
      }
    }
    else
      newImgInfo.imgName = slideName;
    return newImgInfo;
  }

  // initialize panel
  //--------------------------------------------------------------------------
  void initPanel(boolean showButtonPanel)
  {
    setBackground(Color.lightGray);
    setLayout(new BorderLayout());
    sScrPane = new SmartScrollPane(0,100,0,100,0,100,0,100);

    Component thisComp = this; // required in order to make next line compileable in all compilers
    if (thisComp instanceof Panel)
      imgCanvas = new ImgCanvas(imgPanel); // in a Panel we can use heavyweight components
    else
      imgCanvas = new ImgComponent(imgPanel); // to be used in a JPanel

    if (showButtonPanel) {
      addButtonPanel(); // adds a heavyweight toolbar
      // lightweight replacement:
      // ach.JTiffyToolBar toolBar = new ach.JTiffyToolBar();
      // toolBar.setImagePanel(this); // allow toolbar to process events for imagePanel
      // add("North",toolBar);
    }
    else btnSave.setEnabled(saveEnabled);
    if (annotationsEditable)
      addAnnotationPanel();

    imgPanel.setLayout(new BorderLayout());
    imgPanel.add("Center",(Component)imgCanvas);
    add("Center",imgPanel);
    validate();
    ((Component)imgCanvas).requestFocus();
    gcTimer = new GcTimer(8000); // start gc thread with 8 sec cycle time
    gcTimer.start();
  }

  void resizeFrame() {
    boolean isApplet = false;
    Container parent = this;
    Applet aParent = null;
    while (parent.getParent() != null) { // detect top encapsulating parent
      if (parent instanceof java.applet.Applet) // check if there is an applet in the hierarchy
        aParent = (Applet)parent;
      parent = parent.getParent();
    }
    try {
      if (aParent != null)
        isApplet = (aParent.getDocumentBase() != null);
    } catch (Throwable e) { }
    if (!isApplet){ // applets should not be resized
      if (imgSrc != null)
      {
        if (initialSize == null)
          initialSize = parent.getSize(); // assume parent was initially sized to display whole toolbar
        Dimension surroundings = new Dimension(parent.getSize().width - this.getWidth(), parent.getSize().height - this.getHeight());
        int frameWidth = insets().left + insets().right + imgSrc.width + surroundings.width; 
        if (isAncestorOf(catPanel)) { // if required, add size of thumbnails panel
          frameWidth += catPanel.getWidth();
        }
        frameWidth = java.lang.Math.max(frameWidth, initialSize.width);
        int frameHeight = java.lang.Math.max(insets().top  + insets().bottom + imgSrc.height + surroundings.height, initialSize.height);
        Dimension scrSize = getToolkit().getScreenSize();
        // Determine ScreenInsets (supported as of Java 1.4
        // GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
        // GraphicsConfiguration gc = ge.getScreenDevices()[0].getConfigurations()[0];
        // Insets scrInsets = getToolkit().getScreenInsets(gc); // determine size e.g. of Window's taskbar
        // scrSize.setSize(scrSize.getWidth()-scrInsets.left-scrInsets.right, scrSize.getHeight()-scrInsets.top-scrInsets.bottom);
        // Problem: Toolkit of MS VM 32 from SDK 2.0beta always returns 320x240
        // so only use ScreenSize when at least VGA dimensions were returned
        if ((scrSize.width >= 640) && (scrSize.height >= 480))
        {
          if ((scrSize.height % 120 == 0) || (scrSize.height % 50 == 0) || 
              (scrSize.height % 256 == 0)) // 480, 600, 768, 960, 1024, 1050, 1200 etc.
            scrSize.height -= 38; // leave some space for task bar if not detected via screen insets
          frameWidth = Math.min(frameWidth, scrSize.width);
          frameHeight = Math.min(frameHeight, scrSize.height);
          ; // move((scrSize.width-frameWidth) / 2, (scrSize.height - frameHeight) / 2);
          Rectangle b = parent.getBounds();
          frameWidth = Math.min(frameWidth, scrSize.width - b.x);
          frameHeight = Math.min(frameHeight, scrSize.height - b.y);
        }
        parent.resize(frameWidth, frameHeight);
        parent.validate(); // necessary for Java 1.3, 1.4
      }
      
    }
    
  }
  
  // show panel
  //--------------------------------------------------------------------------
  void showPanel()
  {
    if (showCatalog && ((slides != null) || (imgInfo.numPages > 1))) {
      if (slides == null) { // one slide but multipage TIFF
        catSlides = new String[1]; // create catSlides manually
        if (imgInfo.imgURL != null)
          catSlides[0] = imgInfo.imgURL.toString();
        else
          catSlides[0] = imgInfo.imgName;
      }
      if (!isAncestorOf(catPanel)) { // if not already added, do it now
        catPanel = new CatalogPanel(true, null, false);
        add("West", catPanel);
        catPanel.setSize(thumbNailWidth + 29,catPanel.getSize().height);
      }
    }
    tmpImgRect = null;         // initialize rectangle of last operation
    frameUpdated = false;      // force image update
    updatePageNum();
    validate();
    repaint();                 // frameUpdated will be set true if paint method was executed
    if (!annotationMode())
      setCanvasCursor(Cursor.CROSSHAIR_CURSOR);
    else
      setCanvasCursor(Cursor.DEFAULT_CURSOR);
    resizeFrame();
    // if (!frameUpdated) {       // ship around bug of some builds of MS VM
    //   Graphics g = getGraphics();
    //   if (g != null) {
    //     update(g); // force call to paint method
    //   g.dispose();
    //   }
    // }
  }

  /**
   * constructor for ImagePanel (image names passed as URL)
   * @param parent is the container where this component is to be embedded
   * @param URL points to the location of the image to be displayed
   * @param options String with display control options (please refer to TiffyAppletDoc.html)
   * @param showButtonPanel flag controls if the (AWT-based) button panel should be displayd
   */
  public ImagePanel(Container parent, URL context, String options, boolean showButtonPanel)
  {
    super();
    this.parent = parent;
    imgInfo.imgURL = context;
    String name = context.toString();
    if ((name != null) && (name.charAt(name.length() - 1) == '/'))
      name = name.substring(0,name.length() - 1);
    int index, searchLen; // separate image name(s) from URL
    if ((searchLen = name.indexOf(';')) < 0)
      searchLen = name.length();
    int li = name.substring(0,searchLen).lastIndexOf('/');
    if (li < 0)
      li = name.substring(0,searchLen).lastIndexOf('\\'); // \\ is escaped character '\'
    if (li >= 0)
      imgInfo.imgName = name.substring(li + 1);
    parseOptions(options);
    initPanel(showButtonPanel);
    parseSlides();
    loadImage(imgInfo); // asynchronously - calls showPanel() after success
  }

  /**
   * constructor for ImagePanel (image names passed as String)
   * @param parent is the container where this component is to be embedded
   * @param name holds the pathname(s) of the image(s) to be displayed (separated by semicolon)
   * @param options String with display control options (please refer to TiffyAppletDoc.html)
   * @param showButtonPanel flag controls if the (AWT-based) button panel should be displayd
   */
  public ImagePanel(Container parent, String name, String options, boolean showButtonPanel)
  {
    super();
    this.parent = parent;
    imgInfo.imgName = name;
    parseOptions(options);
    initPanel(showButtonPanel);
    parseSlides();
    loadImage(imgInfo); // asynchronously - calls showPanel() after success
  }

  /**
   * constructor for ImagePanel (image passed as ByteArray)
   * @param parent is the container where this component is to be embedded
   * @param imgFileBA holds the ByteArray containing the image file (currently only TIFF, JPG, GIF)
   * @param options String with display control options (please refer to TiffyAppletDoc.html)
   * @param showButtonPanel flag controls if the (AWT-based) button panel should be displayd
   */
  public ImagePanel(Container parent, byte[] imgFileBA, String options, boolean showButtonPanel)
  {
    super();
    this.parent = parent;
    imgInfo.imgName = "";
    imgInfo.imgFileBA = imgFileBA;
    parseOptions(options);
    initPanel(showButtonPanel);
    parseSlides();
    loadImage(imgInfo); // asynchronously - calls showPanel() after success
  }

  /**
   * constructor for ImagePanel (image passed as java.awt.Image)
   * @param parent is the container where this component is to be embedded
   * @param img holds the image to be displayed
   * @param options String with display control options (please refer to TiffyAppletDoc.html)
   * @param showButtonPanel flag controls if the (AWT-based) button panel should be displayd
   */
  public ImagePanel(Container parent, Image img, String options, boolean showButtonPanel)
  {
    super();
    this.parent = parent;
    imgInfo.imgName = "";
    parseOptions(options);
    initPanel(showButtonPanel);
    parseSlides();
    imageProducerUpdate(img.getSource(), null, imgInfo); // calls showPanel() after success
  }

  /**
   * a component specified here will be enabled/disabled depending
   * saving of the image is enabled
   * @param c Component to be controlled
   */
  public void addSaveControl(Component c) {
    cSave = c;
  }

  /**
   * a component specified here will be enabled/disabled depending
   * printing of the image is enabled
   * @param c Component to be controlled
   */
  public void addPrintControl(Component c) {
    cPrint = c;
  }

  /**
   * a component specified here will be enabled/disabled depending
   * printing of the viewable part of the image is enabled
   * @param c Component to be controlled
   */
  public void addPrintViewControl(Component c) {
    cPrintView = c;
  }

  /**
   * a component specified here will be enabled/disabled depending
   * if there is a "previous page" in the list of multiple images
   * @param c Component to be controlled
   */
  public void addPrevPageControl(Component c) {
    cPrevPage = c;
  }

  /**
   * a component specified here will be enabled/disabled depending
   * if there is a "next page" in the list of multiple images
   * @param c Component to be controlled
   */
  public void addNextPageControl(Component c) {
    cNextPage = c;
  }

  /**
   * a component specified here will be enabled/disabled depending
   * if a thumbnail preview is displayed
   * @param c Component to be controlled
   */
  public void addThumbnailsControl(Component c) {
    cThumbnails = c;
    btnThumbnails.mapComponent(c); // allows to control c directly from btnAnnotations
  }

  /**
   * a component specified here will be enabled/disabled depending
   * if annotations exist beside the image
   * @param c Component to be controlled
   */
  public void addAnnotationsControl(Component c) {
    cAnnotations = c;
    btnAnnotations.mapComponent(c); // allows to control c directly from btnAnnotations
  }

  /**
   * a component specified here will be enabled/disabled depending
   * on the state of btnScaleSmooth
   * @param c Component to be controlled
   */
  public void addScaleSmoothControl(Component c) {
    cScaleSmooth = c;
    btnScaleSmooth.mapComponent(c); // allows to control c directly from btnScaleSmooth
  }

  /**
   * Set implementation to be used according to ToolbarController interface
   * @param tbc implementation of ToolbarController interface
   */
  public void setToolbarController(ToolbarController tbc) {
    this.tbc = tbc;
  }

  void setCanvasCursor(int cursorType)
  {
    try
    {  ((Component)imgCanvas).setCursor(new Cursor(cursorType));  }
    // in java version < 1.1 only cursor for frames can be set
    // the kind of error/exception thrown differs between virtual machines
    // so catch the mother of all exceptions
    catch (Throwable e) {
      try { setCursor(new Cursor(cursorType)); } // not supported with embedded applets in 3.x versions of Navigator and MSIE
      catch (Throwable /*SecurityException*/ se) { }
    }
  }

  int getCanvasCursor()
  {
    try
    {  return ((Component)imgCanvas).getCursor().getType();  }
    // in java version < 1.1 only cursor for frames can be got
    // the kind of error/exception thrown differs between virtual machines
    // so catch the mother of all exceptions
    catch (Throwable e) {
      try { return getCursor().getType();  } // not supported with embedded applets in 3.x versions of Navigator and MSIE
      catch (Throwable /*NoSuchMethodError*/ ne) { return 0; }
    }
  }

  // show a modal OK dialog box
  // caution: Java before 1.1 does not pause execution during modal dialog boxes
  void showOkBox(String title, String msg)
  {
    ach.OkBox ok = new ach.OkBox(dummyFrame, title, msg);
    ((Window)ok).show(); // casting to Window required for Sun Java 1.0.2
    ((Component)imgCanvas).requestFocus();
  }

 
  /**
   * Clean up code to free all ressources that may be used by this component
   */
  public void cleanUp()
  {
    // As a modal dialog window opened from a browser applet does not
    // behave as expected, dispose it, if it is still open at cleanup
    if (ach.OkBox.currentOkBox != null) ach.OkBox.currentOkBox.cleanUp();
    boolean isRunningAsApplet = false;
    try { // test if we are running as an applet
      isRunningAsApplet = (((java.applet.Applet)parent).getDocumentBase() != null);
    } catch (Exception e) { }
    if (!isRunningAsApplet) // When running as applet, Sun's Java Plugin will hang MS IE with following popup
      checkAnnotationsModified(); // Give user a chance to save annotations
    props.remove(Annotations.sAnnotations); // remove annotations from properties before saving props
    ParamUtil.saveProps(props, parent);
    dummyFrame.dispose();
    boolean isApplet = false;
    Container parent = this;
    while ((parent.getParent() != null) && !(parent instanceof java.applet.Applet))
      parent = parent.getParent();
    try {
      if ((parent != null) && (parent instanceof java.applet.Applet))
        isApplet = (((java.applet.Applet)parent).getDocumentBase() != null);
    } catch (Throwable e) { }
    gcTimer.shutDown();
    // System.exit() is not allowed for applets
    try {
      if (!isApplet) {
        if (errorMsg == null)
          System.exit(0);
        else
          System.exit(1);
      }
    } catch (Throwable e) { }
  }

  /**
   * show help file with key map
   */
  public void showKeyHelp() {
    System.gc();
    ach.OkBox ok = new ach.OkBox(dummyFrame, DemoLocal.s_keyMap, DemoLocal.s_keyHelpTitle,
      DemoLocal.s_keyHelpInfoText, null, null, null);
    ((Window)ok).show();
    requestFocus();  ((Component)imgCanvas).requestFocus();
  }

  /**
   * show navigation frame ("grabber hand")
   */
  public boolean navigation() {
    if ((navDialog != null) && (navDialog.isShowing()))
      navDialog.toFront();
    else {
      int oldCursor = getCanvasCursor();
      setCanvasCursor(Cursor.WAIT_CURSOR);
      navDialog = new NavDialog(dummyFrame, this, imgSrc, NavDialog.calcFitInDim(new Dimension(256, 256), new Dimension(imgSrc.width, imgSrc.height)));
      navDialog.setViewRect(tmpImgRect);
      setCanvasCursor(oldCursor);
    }
    return true;
  }

  // Get filename using standard FileDialog
  // filenameFilter, masks for filenames and mode (open/save) can be controlled
  // returns null, if no filename selected
  String getFilename(java.io.FilenameFilter filenameFilter, String nameMasks, int mode)
  {
    try
    {
      String imgPath;
      if ((nameMasks.indexOf('*') >= 0) || ((nameMasks.indexOf('/') <= 0) && (nameMasks.indexOf('\\') <= 0)))
          imgPath = imgInfo.imgName;
      else
        imgPath = nameMasks;
      return ParamUtil.getFilename(dummyFrame, filenameFilter, nameMasks, imgPath, mode);
    }
    catch (Throwable e)
    {
      String errStr = e.toString();
      errStr = errStr.substring(errStr.indexOf(':')+1);
      showOkBox(DemoLocal.s_security, DemoLocal.s_appletNoRightFor +  DemoLocal.s_listingFiles + "\r\n" + errStr);
      return null;
    }
  }

  // strip path of a filename that includes a path prefix
  String stripPath(String filename) {
    int li = filename.lastIndexOf('/');
    if (li < 0)
      li = filename.lastIndexOf('\\'); // '\\' is character '\'
    if (li >= 0)
      return filename.substring(li + 1);
    else
      return filename;
  }

  /**
   * Toggle display of thumbnail preview
   */
  public void toggleThumbnails() {
    if (imgInfo.numPages > 1) {
      if (slides == null) { // one slide but multipage TIFF
        catSlides = new String[1]; // create catSlides manually
        if (imgInfo.imgURL != null)
          catSlides[0] = imgInfo.imgURL.toString();
        else
          catSlides[0] = imgInfo.imgName;
      }
    }
    if (!isAncestorOf(catPanel)) { // if not already added, do it now
      showCatalog = true;
      if (catPanel == null)
        catPanel = new CatalogPanel(true, null, false);
      add("West", catPanel);
      catPanel.setSize(thumbNailWidth + 29,catPanel.getSize().height);
    }
    else {
      remove(catPanel);
      showCatalog = false;
    }
    validate();
    repaint();
  }
  
  void enableAnnotations(ShapeList shapeList) {
    graphEd.setShapeList(shapeList);
    showAnnotations = true;
    btnAnnotations.setToggleState(true); // set toggle button in pushed state
    if (annotationsEditable) {
      if (!this.isAncestorOf(annotationPanel.getParent())) {
        add("South", annotationPanel.getParent()); // annotationPanel is embedded in another panel
        if (btnPointer.getToggleState())
          setCanvasCursor(Cursor.DEFAULT_CURSOR);
      }
    }
  }

  void disableAnnotations() {
    if (annotationsEditable)
      remove(annotationPanel.getParent()); // annotationPanel is embedded in another panel
    showAnnotations = false;
    btnAnnotations.setToggleState(false); // set toggle button in not-pushed state
    // graphEd = null;  graphEd = new GraphicEditor(this);
    graphEd.setShapeList(null);
    annotationButtonGroup.setSelected(btnPointer, true);
    setGraphEdMode(GraphicEditor.POINTERMODE);
  }

  /**
   * Get currently loaded / edited annotations.
   * @param resetModifiedState flag. If true, all modified flags will be reset to unmodified.
   * @return String with an XML representation of the annotations or null if no annotations exist.
   */
  public String getAnnotations(boolean resetModifiedState) {
    if (resetModifiedState)
      for (int i=0; i<imgInfo.shapeLists.length; i++)
        imgInfo.shapeLists[i].resetModifiedState();
    return Annotations.getXMLRepresentation(imgInfo.shapeLists, "name=\"" + imgInfo.imgName + "\"");
  }

  /**
   * Set (or replace current) annotations and refresh display.
   * If image is still loading asynchronously, execution is scheduled
   * for next repaint event.
   * @param annotations String containing an XML representation of the annotations.
   */
  public void setAnnotations(String annotations) {
    bufferedAnnotations = new String(annotations); // create a copy
    forceImageUpdate(); // further processing will take place at repaint
  }

  /**
   * Replace default AnnotationSaver implementation with your own implementation.
   * Method saveAnnotations() is called when user clicked the "Save annotions" button.
   * @param AnnotationSaver an instance of a class implementing ach.vectorGraphics.AnnotationSaver interface.
   */
  public void setAnnotationSaver(AnnotationSaver as) {
    annotationSaver = as;
  }

  /**
   * Default implementation of the AnnotationSaver interface to save current
   * annotations. Can be replaced by your own implementation by
   * setAnnotionSaver(). Method saveAnnotations() of currently set AnnotationSaver
   * is called when user clicked the "Save annotions" button or confirms the
   * "save annotations" popup after annotations have been modified.
   */
  public void saveAnnotations() {
    if ((imgInfo.shapeLists == null) || (imgInfo.shapeLists.length == 0))
      return;
    String name = stripPath(imgInfo.imgName);
    int li = name.lastIndexOf(".");
    if (li >=0)
      name = name.substring(0,li);
    name = name + imgInfo.annotationExtension;
    name = getFilename(new ach.vectorGraphics.XMLFilenameFilter(), name, FileDialog.SAVE);
    if ((name == null) || (name == ""))
      return;
    try {
      FileWriter fw = new FileWriter(name);
      fw.write(getAnnotations(true));
      fw.close();
      for (int i=0; i<imgInfo.shapeLists.length; i++)
        imgInfo.shapeLists[i].resetModifiedState();
    } catch (Exception e) {
      showOkBox(DemoLocal.s_error, e.toString() + ": " + name);
      return;
    }
  }

  // Downloads the image currently loaded to a local file
  void downloadImage() {
    if (imgSrc == null)
      return;
    String name = getFilename(null, stripPath(imgInfo.imgName), FileDialog.SAVE);
    if ((name == null) || (name == ""))
      return;
    URL context = imgInfo.imgURL;
    try {
      if (context == null)
        context = new URL("file:" + imgInfo.imgName); // may throw java.net.MalformedURLException
      ParamUtil.netFileToLocalFile(context, name, true); // may throw java.io.IOException
    } catch (Exception e) {
      showOkBox(DemoLocal.s_error, e.toString() + ": " + context.toString());
      return;
    }
  }

  /**
   * Implementation of Scrollable interface
   * @param x absolute x-position of upper left edge of viewable rectangle
   * @param y absolute y-position of upper left edge of viewable rectangle
   */
  public void scrollAbs(int x, int y) {
    sScrPane.sbHoriz.setValue((int)(x * xZoom));
    sScrPane.sbVert.setValue((int)(y * yZoom));
    sScrPane.saveValues();
    repaint();
  }

  // adjust step-size and suppress echos of scrollbar events
  // depending on java version and VM supplier
  // finally schedule a repaint
  //--------------------------------------------------------------------------
  synchronized void doScrollStep(Object target, int direction, boolean pageStep) {
    sScrPane.doScrollStep(target, this, direction, pageStep);
    repaint();
  }

  // Get current image offset controlled by scrollbars
  Point imageOffset() {
    Point p = new Point(sScrPane.getHorizValue(imgPanel,(int)Math.round(xZoom * imgSrc.width)),
                        sScrPane.getVertValue(imgPanel,(int)Math.round(yZoom * imgSrc.height)));
    return p;
  }

  // translate virtual (scaled and shifted) selected area (r)
  // to actual image coordinates of selected area
  Rectangle getRealCoordinates(Rectangle r) {
    Point imgOffset = imageOffset();
    return new Rectangle(
      (int)Math.round((float)(imgOffset.x + r.x)/xZoom),
      (int)Math.round((float)(imgOffset.y + r.y)/yZoom),
      (int)Math.round(Math.floor((float)r.width/xZoom)),
      (int)Math.round(Math.floor((float)r.height/yZoom)));
  }

  // Calulate a new image offset based on an old image offset and current zoom
  // factor for an image that is centered at the same point as before.
  // Set the scrollbar values to the new image offset
  void setNewImageOffset(double oldxZoom, double oldyZoom, Point oldImgOffset) {
    int imgWidth = imgSrc.width;
    int imgHeight = imgSrc.height;
    int halfVisiWidth = sScrPane.getVisibleWidth(this) / 2;
    int halfVisiHeight = sScrPane.getVisibleHeight(this) / 2;
    int hValue = (int)Math.round(xZoom *
                    (oldImgOffset.x  + halfVisiWidth) / oldxZoom)
                    - halfVisiWidth;
    int vValue = (int)Math.round(yZoom *
                    (oldImgOffset.y + halfVisiHeight) / oldyZoom)
                    - halfVisiHeight;
    Dimension newZoomedImg = new Dimension((int)Math.round(xZoom*imgWidth),
                                           (int)Math.round(yZoom*imgHeight));
    sScrPane.setHorizValue(hValue, this, newZoomedImg.width);
    sScrPane.setVertValue(vValue, this, newZoomedImg.height);
  }

  /**
   * increase zoom factor and calulate new image offset for centered image
   */
  public void zoomIn() {
    fitHoriz = false;
    fitInWindow = false;
    Point oldImgOffset = imageOffset();
    double oldxZoom = xZoom,
           oldyZoom = yZoom;
    if (xZoom >= 1)
      xZoom = Math.round(xZoom + 1);
    else if (xZoom > 0.5)
      xZoom = 1;
    else
      xZoom = (double)1 / Math.round(1 / xZoom - 1);
    yZoom = xZoom;
    setNewImageOffset(oldxZoom, oldyZoom, oldImgOffset);
    repaint();
  }

  /**
   * decrease zoom factor and calulate new image offset for centered image
   */
  public void zoomOut() {
    fitHoriz = false;
    fitInWindow = false;
    Point oldImgOffset = imageOffset();
    double oldxZoom = xZoom,
           oldyZoom = yZoom;
    if (Math.round(xZoom) >= 2)
      xZoom = Math.round(xZoom - 1);
    else
      xZoom = (double)1 / Math.round(1 / xZoom + 1);
    yZoom = xZoom;
    setNewImageOffset(oldxZoom, oldyZoom, oldImgOffset);
    repaint();
  }

  /**
   * set free defined zoom factor and calulate new image offset for centered image
   */
  public void zoom(double zoom) {
    if ((xZoom != zoom) || (yZoom != zoom))
    {
      fitHoriz = false;
      fitInWindow = false;
      xZoom = zoom;  yZoom = zoom;
      repaint();
    }
  }

  /**
   * set 1:1 view (zoom factor 1) and calulate new image offset for centered image
   */
  public void zoom1To1() {
    zoom(1);
  }

  /**
   * scale image to fit into the window horizontally
   */
  public void fitHoriz() {
    if (!fitHoriz) {
      fitHoriz = true;
      fitInWindow = false;
      repaint();
    }
  }

  /**
   * scale image to fit into the window completely
   */
  public void fitInWindow() {
    if (!fitInWindow) {
      fitInWindow = true;
      fitHoriz = false;
      repaint();
    }
  }

  // calculate zoom factor and image offsets to display given rectangle
  void zoomToRect(Rectangle r) {
    if ((r.width == 0) || (r.height == 0))
      return;
    fitHoriz = false;
    fitInWindow = false;
    Point imgOffset = imageOffset();
    double oldxZoom = xZoom,
           oldyZoom = yZoom;
    // calulate new zoom factor
    xZoom = xZoom * sScrPane.getVisibleWidth(this) / r.width;
    yZoom = yZoom * sScrPane.getVisibleHeight(this) / r.height;
    if (xZoom > yZoom)  xZoom = yZoom;
    else yZoom = xZoom;
    // calculate new values for scrollbars (as offset)
    int hValue = (int)Math.round(xZoom / oldxZoom * (imgOffset.x + r.x));
    int vValue = (int)Math.round(yZoom / oldyZoom * (imgOffset.y + r.y));
    int imgWidth = imgSrc.width;
    int imgHeight = imgSrc.height;
    Dimension newZoomedImg = new Dimension((int)Math.round(xZoom*imgWidth),
                                           (int)Math.round(yZoom*imgHeight));
    sScrPane.setHorizValue(hValue, this, newZoomedImg.width);
    sScrPane.setVertValue(vValue, this, newZoomedImg.height);
    repaint();
  }

  /**
   * toggle smooth scaling (antialiasing)
   */
  public void toggleSmoothScaling() {
    if (imgSrc.calcPixelSize() == 1) {
      showOkBox("Antialiasing", DemoLocal.s_scaleSmooth + " " + DemoLocal.s_activatedByDefault + "\r\n" +
        DemoLocal.s_forBilevelImages);
      return;
    }
    if (!scaledInstanceSupported) {
      showOkBox("Antialiasing", DemoLocal.s_scaleSmooth + DemoLocal.s_notSupportedByVM + "\r\n" +
        DemoLocal.s_java11required);
      return;
    }
    if (scaleHint == Image.SCALE_FAST) {
      scaleHint = Image.SCALE_SMOOTH;
      btnScaleSmooth.setBackground(Color.gray);
    } else {
      scaleHint = Image.SCALE_FAST;
      btnScaleSmooth.setBackground(Color.lightGray);
    }
    tmpImgRect = null; // force image update
    repaint();
  }

  void forceImageUpdate() {
    tmpImgRect = null;
    repaint();
  }

  /**
   * toggle display of annotations
   */
  public void toggleAnnotations() {
    if (btnAnnotations.isEnabled()) {
      showAnnotations = !showAnnotations;
      if (annotationsEditable) {
        if (showAnnotations) {
          add("South", annotationPanel.getParent()); // annotationPanel is embedded in another panel
          if (btnPointer.getToggleState())
            setCanvasCursor(Cursor.DEFAULT_CURSOR);
          if (imgInfo.shapeLists == null) {
            imgInfo.shapeLists = new ShapeList[imgInfo.numPages];
            for (int i = 0; i < imgInfo.numPages; i++)
              imgInfo.shapeLists[i] = new ShapeList();
            graphEd.setShapeList(imgInfo.shapeLists[imgInfo.page-1]);
          }
        }
        else
          if (this.isAncestorOf(annotationPanel.getParent())) {
            remove(annotationPanel.getParent()); // annotationPanel is embedded in another panel
            setCanvasCursor(Cursor.CROSSHAIR_CURSOR);
          }
      }
      validate();
      forceImageUpdate();
    }
  }

  /**
   * for multi-document images step to indicated page
   * for a slide sequence set slide index to indicated slide
   * @param pageNum index of targeted page; counting starts with 1
   */
  public boolean setPageNum(int pageNum) {
    if (pageNum < 1)
      pageNum = 1;
    if ((slides != null) && (pageNum > slides.length + imgInfo.numPages - 1))
      pageNum = slides.length + imgInfo.numPages - 1;
    if ((slides == null) && (pageNum > imgInfo.numPages))
      pageNum = imgInfo.numPages;
    if (slides != null) {
      if ((slides.length >= pageNum) && (slides[pageNum-1].endsWith("]") || slides[slideIdx].endsWith("]"))) {
        // slide list was updated for multipage TIFFs - all pages can be directly adressed
        slideIdx = pageNum - 1;
        ImageInfo newImgInfo = slideToImageInfo(); // load imageInfo with settings for indicated slide
        loadImage(newImgInfo); // asynchronously - calls showPanel() after success
      } else { // try to work it out in a mixture of single page slides and multipage TIFFs...
        if (slideIdx + imgInfo.page != pageNum) {
          if ((slideIdx + imgInfo.numPages < pageNum) || (pageNum - slideIdx < 1)) {
            slideIdx = pageNum - imgInfo.numPages;
            pageNum -= (imgInfo.numPages - 1); // necessary for correct printing of sequences of multipage TIFFs
            printPageNum -= (imgInfo.numPages - 1); // necessary for correct printing of sequences of multipage TIFFs
            imgInfo.page = 1;
            if (slides.length >= pageNum) {
              slideIdx = pageNum - 1;
              imgInfo.page = 1;
            } else {
              slideIdx = slides.length - 1;
              imgInfo.page = pageNum - slides.length;
            }
          }
          ImageInfo newImgInfo = slideToImageInfo(); // load imageInfo with settings for indicated slide
          newImgInfo.page = pageNum - slideIdx;
          loadImage(newImgInfo); // asynchronously - calls showPanel() after success
        }
      }
    } else {
      if (imgInfo.page != pageNum) {
        imgInfo.page = pageNum;
        loadImage(imgInfo); // asynchronously - calls showPanel() after success
      }
    }
    return true;
  }

  /**
   * for multi-document images step back to previous page
   * for a slide sequence decrement slide index
   */
  public boolean setPreviousPage() {
    if (imgInfo.page > 1) {
      imgInfo.page--;
      loadImage(imgInfo); // asynchronously - calls showPanel() after success
      if ((slides != null) && (slides[slideIdx].endsWith("]")))
        slideIdx--; // slide list was updated for direct multipage access
    }
    else if ((slides != null) && (slideIdx > 0)) {
      slideIdx--;
      if (slides[slideIdx] != null) {
        ImageInfo newImgInfo = slideToImageInfo(); // load imageInfo with settings for previous slide
        loadImage(newImgInfo); // asynchronously - calls showPanel() after success
      } else slideIdx = 0;
    }
    return true;
  }

  /**
   * for multi-document images step forward to next page
   * for a slide sequence increment slide index
   */
  public boolean setNextPage() {
    if (imgInfo.page < imgInfo.numPages) {
      imgInfo.page++;
      loadImage(imgInfo); // asynchronously - calls showPanel() after success
      if (slides != null) {
        if (slideIdx < 0)
          slideIdx = 0;
        if (slides[slideIdx].endsWith("]"))
          slideIdx++; // slide list was updated for direct multipage access
      }
    }
    else if ((slides != null) && (slideIdx < slides.length-1)) {
      slideIdx++;
      if (slides[slideIdx] != null) {
        ImageInfo newImgInfo = slideToImageInfo();
        if (!slides[slideIdx].endsWith("]"))
          newImgInfo.page = 1;
        loadImage(newImgInfo); // asynchronously - calls showPanel() after success
      } else slideIdx = slides.length-1;
    }
    else if (printJob != null) {
      if (printJob instanceof PrintJob)
        ((PrintJob)printJob).end();
      else {
        PrintImage2.printBook(printJob, book);
        book = null;
      }
      printJob = null;
    }
    return true;
  }

  /**
   * ToolbarController: set text of PageNum TextField to page number
   */
  public void setPageNumText(int pageNum) {
    tfPageNum.setText(String.valueOf(pageNum));
  }

  /**
   * ToolbarController: get text of PageNum TextField converted to int
   */
  public int getPageNum() {
    return Integer.parseInt(tfPageNum.getText());
  }

  /**
   * ToolbarController: set text of numPages Label to number of pages
   */
  public void setNumPagesText(int numPages) {
    lNumPages.setText(" / " + String.valueOf(numPages));
  }

  /**
   * ToolbarController: set text of zoom TextField to zoomFactor
   * (update only if current text is different to avoid unnecessary repaints)
   */
  public void setZoomText(double zoomFactor) {
    String zoomText = Long.toString(java.lang.Math.round(zoomFactor*100)) + "%";
    if (!tfZoom.getText().equals(zoomText))
      tfZoom.setText(zoomText);
  }

  void getSelected() {
    if (showCatalog) {
      String[] selectedSlides = catPanel.getSelected();
      if ((selectedSlides != null) && (selectedSlides.length > 0)) {
        String firstSlide = selectedSlides[0]; // car only for first selected slide
        int page = 1;
        while ((page <= slides.length) && (!slides[page-1].equals(firstSlide)))
          page++;
        setPageNum(page);
      }
    }
  }

  /**
   * ToolbarController: get zoom factor from zoom TextField
   */
  public double getZoomFactor(double defaultZoom) {
    String zoomStr = tfZoom.getText();
    double zoomFactor = defaultZoom;
    try {
      if (zoomStr.indexOf('%') >= 0)
        zoomStr = zoomStr.substring(0,zoomStr.indexOf('%'));
      zoomFactor = Double.valueOf(zoomStr).doubleValue() / 100; // some VMs don't accept Double.parseDouble()
    } catch (Exception e) {  }
    return zoomFactor;
  }

  /**
   * update pageNum text field and numpages label
   * for a slide sequence set pageNum to slide index
   * for a multi-document image set pageNum to page index
   */
  public void updatePageNum() {
    int pageNum = 0;
    int numPages = 0;
    if (imgInfo != null) {
      pageNum = imgInfo.page;
      numPages = imgInfo.numPages;
    }
    if (slides != null) {
      if (slides[slideIdx].endsWith("]")) {
        pageNum = slideIdx + 1;
        numPages = slides.length;
      }
      else {
        pageNum += slideIdx;
        numPages += slides.length - 1;
      }
    }
    tbc.setPageNumText(pageNum);
    tbc.setNumPagesText(numPages);
    if (numPages == 1)
      tfPageNum.disable();
    else
      tfPageNum.enable();
    if (showCatalog && isAncestorOf(catPanel)) {
      catPanel.setSelectedAll(false);
      catPanel.setSelected(slideIdx, true);
    }
  }

  /**
   * close navigation dialog (if displayed)
   */
  void closeNavDialog() {
    if (navDialog != null) {
      navDialog.dispose();
      navDialog = null;
    }
  }

  boolean annotationMode() {
    return (btnAnnotations.isEnabled() && showAnnotations);
  }

  void setGraphEdMode(int mode) {
    Graphics g = getGraphics();
    graphEd.setMode(mode, g);
    update(g);
    g.dispose();
  }

  void applyShapeListRotation(int rotationIndex) {
    if ((imgInfo.shapeLists != null) && (imgInfo.shapeLists[imgInfo.page-1] != null))
      imgInfo.shapeLists[imgInfo.page-1].applyQuadrantRotation(rotationIndex);
  }

  /**
   * rotate image by 90 degree
   */
  public boolean rotate090() {
    int oldCursor = getCanvasCursor();
    setCanvasCursor(Cursor.WAIT_CURSOR);
    EMemImageSource tmpImgSrc =  imgSrc.getRotated090();
    imgSrc = tmpImgSrc;
    applyShapeListRotation(1);
    forceImageUpdate();
    setCanvasCursor(oldCursor);
    closeNavDialog();
    return true;
  }

  /**
   * rotate image by 180 degree
   */
  public boolean rotate180() {
    int oldCursor = getCanvasCursor();
    setCanvasCursor(Cursor.WAIT_CURSOR);
    imgSrc.rotate180();
    applyShapeListRotation(2);
    forceImageUpdate(); // force image update although bound of visible area did not change
    setCanvasCursor(oldCursor);
    return true;
  }

  /**
   * rotate image by 270 degree
   */
  public boolean rotate270() {
    int oldCursor = getCanvasCursor();
    setCanvasCursor(Cursor.WAIT_CURSOR);
    EMemImageSource tmpImgSrc =  imgSrc.getRotated270();
    imgSrc = tmpImgSrc;
    applyShapeListRotation(3);
    forceImageUpdate(); // force image update although bound of visible area did not change
    setCanvasCursor(oldCursor);
    closeNavDialog();
    return true;
  }

  /**
   * mirror image horizontally
   */
  public boolean mirror() {
    int oldCursor = getCanvasCursor();
    setCanvasCursor(Cursor.WAIT_CURSOR);
    imgSrc.mirrorHorizontal();
    forceImageUpdate(); // force image update although bound of visible area did not change
    setCanvasCursor(oldCursor);
    closeNavDialog();
    return true;
  }

  /**
   * flip image vertically
   */
  public boolean flip() {
    int oldCursor = getCanvasCursor();
    setCanvasCursor(Cursor.WAIT_CURSOR);
    imgSrc.flipVertical();
    forceImageUpdate(); // force image update although bound of visible area did not change
    setCanvasCursor(oldCursor);
    closeNavDialog();
    return true;
  }

  /**
   * invert image
   */
  public boolean invert() {
    imgSrc.invert(); // EMemImageSource has its own inversion method
    forceImageUpdate(); // force image update although bound of visible area did not change
    return true;
  }


  /**
   * Open print dialog and start printing.
   * @param imgSrc is the image producer (EMemImageSource) to be printed
   * @param asynchronousFlag controls whether printing should be carried
   * out as asynchronous thread or synchronously
   * Another more detailed example to use PrintDialog and PrintImage see method printOverlayedImage.
   */
  boolean printImage(EMemImageSource imgSrc, boolean asynchronousFlag) {
    if (quickPrint){
      if (printJob == null) {
        printJob = getToolkit().getPrintJob(dummyFrame, stripPath(imgInfo.imgName), null);
        printPageSelection = "ALL";
        printPageNum = nextPageFromPrintPageSelection(printPageSelection, 0);
        // quickPrint = false;
      }
      if (printJob != null) {
        try {
          // due to precompilation of some code that will never be executed in MS IE...
          autoPrintImage(printJob); // ...currently causes an IOException in MS IE (trying to read PrinterJob)
        } catch (Throwable e) { }
        printPageNum = nextPageFromPrintPageSelection(printPageSelection, printPageNum);
        if (printPageNum >= 0)
          setPageNum(printPageNum);
        else { // printPageNum < 0 signals end of job
          if (printJob instanceof PrintJob)
            ((PrintJob)printJob).end();
          else {
            PrintImage2.printBook(printJob, book);
            book = null;
          }
          printJob = null;
        }
      }
    } else {
      String pageRange = "";
      if (((imgInfo.numPages > 1) || ((slides != null) && (slides.length > 1))) &&
          (imgSrc == this.imgSrc))
        pageRange = "ALL"; // make "all pages" the default selection in printer dialog
      props.put("PAGERANGE",pageRange);
      props.put("AUTOROTATEALLOWED", new Boolean(autorotateAllowed));
      // props.put("RESOLUTION","150 dpi"); // may be used to set the default resolution in MS IE to 150 dpi
      if (annotationMode()) {
        props.put(Annotations.sAnnotations,imgInfo.shapeLists[imgInfo.page-1]);
      } else
        props.remove(Annotations.sAnnotations);
      try {
        PrintImage pi = new PrintImage(dummyFrame, imgSrc, props, (BatchPrinter)this);
        if (asynchronousFlag)
          new Thread(pi).start();
        else
          pi.run();
        // if lightweight components are used, it might be necessary to use
        // the invokeLater-method instead of Thread(...).start()
        // but this is no general rule...
        // javax.swing.SwingUtilities.invokeLater(new Thread(pi));
      } catch (NoClassDefFoundError e) {
        showOkBox(DemoLocal.s_Printing, DemoLocal.s_eventModel + DemoLocal.s_notSupportedByVM + "\r\n" + DemoLocal.s_java11required);
      } catch (Throwable e) {
        String errMsg = e.toString();
        showOkBox(DemoLocal.s_Printing, errMsg);
      }
      ((Component)imgCanvas).requestFocus();
    }
    return true;
  }

  /**
   * Open print dialog and start printing the current image
   * @param asynchronousFlag controls whether printing should be carried
   * out as asynchronous thread or synchronously
   * Another more detailed example to use PrintDialog and PrintImage see method printOverlayedImage.
   */
  public boolean printImage(boolean asynchronousFlag) {
    return printImage(imgSrc, asynchronousFlag);
  }

  /**
   * Open print dialog and start printing the current viisible area of the image
   * @param asynchronousFlag controls whether printing should be carried
   * out as asynchronous thread or synchronously
   * Another more detailed example to use PrintDialog and PrintImage see method printOverlayedImage.
   */
  public boolean printCurrentView(boolean asynchronousFlag) {
    Rectangle r = newTmpImgRect;
    return printImage(imgSrc.extractMemImgRegion(r.x, r.y, r.width, r.height), asynchronousFlag);
  }

  // get an integervalue from properties-hashtable
  // if no property for key was found return def as default value
  int getIntProp(String key, int def) {
    Integer i = (Integer)props.get(key);
    if (i != null)
      return i.intValue();
    else
      return def;
  }

  /**
   * Print current image fitted into current printer's page dimension
   * reported by the operating system minus margins stored in
   * the properties "LEFTMARGIN", "RIGHTMARGIN", "TOPMARGIN" and
   * "BOTTOMMARGIN" as integer values (1 unit = 1 cm / 1000 = 1 inch / 2540)
   * Print to a given instance of a PrintJob or PrinterJob
   * @param printJob instance of PrintJob or PrinterJob
   */
  boolean autoPrintImage(Object printJob)
  {
    if (printJob != null) {
      Image img = parent.createImage(imgSrc);
      Dimension d;
      if (printJob instanceof PrintJob)
        d = ((PrintJob)printJob).getPageDimension();
      else {
        d = PrintImage2.getPageDimension(printJob);
      }
      int res = 72; // default resolution for printing measurements
      if (printJob instanceof PrintJob) {
        res = preferredResolution;
        if (res == 0) res = ((PrintJob)printJob).getPageResolution();
      }
      int inchFak = 2540; // margin units per inch
      // calculate printable width and height taking margins into account
      // if no margin properties defined, assume 0.5 cm each (1 cm bottom)
      int lMargin = getIntProp("LEFTMARGIN",500) * res / inchFak;
      int hMargins = lMargin + getIntProp("RIGHTMARGIN",500) * res / inchFak;
      int tMargin = getIntProp("TOPMARGIN",500) * res / inchFak;
      int vMargins = tMargin + getIntProp("BOTTOMMARGIN",1000) * res / inchFak;
      if (d.width > d.height) { // printer set to landscape format
        // swap margins
        int temp = hMargins; hMargins = vMargins; vMargins = temp;
        temp = lMargin; lMargin = tMargin; tMargin = temp;
      }
      int pWidth = d.width - hMargins;
      int pHeight = d.height - vMargins;
      if (autorotateAllowed && (((pHeight > pWidth) && (imgSrc.width > imgSrc.height)) || // paper portrait but image landscape
          ((pHeight < pWidth) && (imgSrc.width < imgSrc.height)))) {  // paper landscape but image portrait
        imgSrc = imgSrc.getRotated270(); // rotate to better fit inside paper orientation
        applyShapeListRotation(3);
        img = createImage(imgSrc);
      }
      // calculate optimal pWidth and pHeight according to image proportions
      if ((long)pWidth * imgSrc.height / imgSrc.width > pHeight)
        pWidth = (int)((long)pHeight * imgSrc.width / imgSrc.height);
      else
        pHeight = (int)((long)pWidth * imgSrc.height / imgSrc.width);
      // calculate offsets for image centered on page
      int pX = (d.width - hMargins - pWidth) / 2 + lMargin;
      int pY = (d.height - vMargins - pHeight) / 2 + tMargin;
      if (!(printJob instanceof PrintJob)) { // either PrinterJob of Java >= 1.2 or WPrinterJob of MS JVM
        // rescale image to meet preferred print resolution
        int newWidth = pWidth * preferredResolution / 72;
        int newHeight = pHeight * preferredResolution / 72;
        if (newWidth < imgSrc.width) { // rescale image to lower dimension
          imgSrc = new EMemImageSource(imgSrc.extractMemImgRegion(0,0,imgSrc.width,imgSrc.height,newWidth,newHeight), true);
          img = createImage(imgSrc);
        }
      }
      if (annotationMode()) {
        props.put(Annotations.sAnnotations,imgInfo.shapeLists[imgInfo.page-1]);
      } else
        props.remove(Annotations.sAnnotations);
      try {
        if (printJob instanceof PrintJob) {
          PrintJob pj = (PrintJob)printJob; // make object to what it really is
          Graphics pg;
          pg = pj.getGraphics();
          if (pg != null) {
            // imgSrc.useProdGauge = true; // show gauge during image production
            PrintImage.rPD(pj);
            pg.drawImage(img, pX, pY, pWidth, pHeight, this);
            if (btnAnnotations.isEnabled() && showAnnotations) {
              imgInfo.shapeLists[imgInfo.page-1].applyBounds(new Rectangle(pX, pY, pWidth, pHeight));
              imgInfo.shapeLists[imgInfo.page-1].drawAll(pg, false);
            }
            // imgSrc.useProdGauge = false;
            pg.dispose(); // flush page
          }
        }
        else // if printJob instanceof PrinterJob (Java 2)
          // if book is initially null, printImage will instantiate it to class Book
          book = PrintImage2.printImage(printJob, book, img, imgSrc.width, imgSrc.height, new Dimension(pWidth, pHeight), pX, pY, pWidth, pHeight, props);
      } catch (Throwable e) {
        showOkBox(DemoLocal.s_Printing, e.toString());
      }
      System.gc();
      return true;
    } else
      return false; // no PrintJob given
  }

  /**
   * Determine next page number from a page numbers specification
   * Example: Consecutive calls for "1;3;5-7" will return 1, 3, 5, 6, 7, -1
   * @param pageRange specifies semicolon seperated pagenumbers or ranges
   * @param currentPage specifies current page number
   * @return vaid page number that is bigger than currentpage or -1
   */
  int nextPageFromPrintPageSelection(String printPageSelection, int currentPage) {
    int nextNum = -1;
    String nextNumStr = "";
    StringTokenizer st = new StringTokenizer(printPageSelection, " ;,");
    while (st.hasMoreTokens() && (nextNum <= currentPage)) {
      try {
        nextNumStr = st.nextToken();
        if (nextNumStr.equalsIgnoreCase("ALL"))
          nextNum = currentPage+1;
        else
          nextNum = Integer.parseInt(nextNumStr);
      }
      catch (NumberFormatException nfe) {
        try {
          StringTokenizer st2 = new StringTokenizer(nextNumStr, " -");
          int numLow = Integer.parseInt(st2.nextToken());
          int numHigh = Integer.parseInt(st2.nextToken());
          if (numLow > currentPage)
            nextNum = numLow;
          else if (numHigh > currentPage)
            nextNum = currentPage + 1;
        } catch (Exception ex) { // Either not 2 tokens or no integers
        }
      }
    }
    int numPages = imgInfo.numPages;
    if (slides != null)  {
      if (showCatalog && isAncestorOf(catPanel))
        numPages = slides.length; // if catpanel is displayed, slides.length is total count (including multipage images)
      else
        numPages += slides.length - 1;
    }
    if ((nextNum > numPages) || (nextNum <= currentPage))
      nextNum = -1;
    return nextNum;
  }

  /**
   * Print all slides of a slide sequence and/or all pages of a multipage TIFF
   * either to a PrintJob (JDK 1.1) or PrinterJob (Java 2) (one of them must be not null).
   * Images are automatically scaled proportionally to fit into the printers
   * page size reported by the operating system minus margins stored in
   * the props (hashtable with properties) "LEFTMARGIN", "RIGHTMARGIN", "TOPMARGIN" and
   * "BOTTOMMARGIN" as integer values (1 unit = 1 cm / 1000 = 1 inch / 2540)
   * @param printJob instance of PrintJob (e.g. instantiated by PrintDialog) or PrinterJob
   * @param preferredResolution specifies the preferred print resolution (only used for MS IE)
   * @param pageRange page numbers seperated by semicolons and low - high ranges like 1;2;8-10
   */
  public void batchPrint(Object printJob, int preferredResolution, String printPageSelection)
  {
    if (printJob != null) {
      this.printJob = printJob;
      this.preferredResolution = preferredResolution;
      int pageNum = imgInfo.page;
      if (slides != null)
        pageNum += slideIdx;
      this.printPageSelection = printPageSelection;
      printPageNum = nextPageFromPrintPageSelection(printPageSelection, 0);
      if (pageNum == printPageNum) {
        autoPrintImage(printJob);
        printPageNum = nextPageFromPrintPageSelection(printPageSelection, printPageNum);
        if (printPageNum >= 0)
          setPageNum(printPageNum);
        else {
          if (printJob instanceof PrintJob)
            ((PrintJob)printJob).end();
          else {
            PrintImage2.printBook(printJob, book);
            book = null;
          }
          printJob = null;
        }
      }
      else
        setPageNum(printPageNum); // printing and stepping to next page initiated through callback-method imageProducerUpdate
    }
  }

  /**
   * Method demonstrating the usage of class PrintDialog and PrintImage.
   * In this demonstration a PrintDialog allows to specify image size and
   * location for the main image. During printing second image will be
   * "stamped" as overlay on top of the main image.
   * If no Rectangle and EMemImageSource is specified (handed over as null)
   * this demonstration will stamp the registration dialog image on top.
   * @param printArea2 Rectangle specifying size and location of the overlay image
   * @param imgSrc2 EMemImageSource for the overlay image
   */
  /*
  public boolean printOverlayedImage(Rectangle printArea2, EMemImageSource imgSrc2)
  {
    if (!versionEqualGreater1_1) {
      showOkBox(DemoLocal.s_Printing , DemoLocal.s_Printing + DemoLocal.s_notSupportedByVM + "\r\n" +
        DemoLocal.s_java11required);
    }
    PrintJob printJob = null;
    try {
      try { // Requesting privileges for Netscape Communicator
        if (Class.forName("netscape.security.PrivilegeManager") != null)
          netscape.security.PrivilegeManager.enablePrivilege("UniversalPrintJobAccess");
      } catch( Throwable ge ) { // ge.printStackTrace();
      }
      try { // assert permission for printing from MS Internet Explorer
        if (Class.forName("com.ms.security.PolicyEngine") != null) {
          com.ms.security.PolicyEngine.assertPermission(
            com.ms.security.PermissionID.PRINTING);
        }
      } catch (Throwable cnfe) {  }
      printJob = getToolkit().getPrintJob(dummyFrame, "Bitmap printout", null);
    } catch (Throwable e) {
        showOkBox(DemoLocal.s_security, e.toString() + "\r\n" + DemoLocal.s_appletNoRightFor + DemoLocal.s_Printing + "?");
    }
    Rectangle printArea = null;
    EMemImageSource imgSrc2 = null;
    try {
      if (imgSrc2 == null) {
        imgSrc2 = new EMemImageSource(getToolkit().createImage(
          ParamUtil.getResourceAsByteArray("vibrio.gif",
          Class.forName("ach.file.ParamUtil"))).getSource(), false);
      }
      if (printArea2 == null) // locate stamp close to upper left corner
        printArea2 = new Rectangle(15, 15, imgSrc2.width/2, imgSrc2.height/2);
    } catch (Exception e) { showOkBox(DemoLocal.s_Printing, DemoLocal.s_fileNotFoundOrError); }
    if (printJob != null) {
      int screenRes = parent.getToolkit().getScreenResolution();
      Image img = parent.createImage(imgSrc);
      Dimension d = printJob.getPageDimension();
      int res = printJob.getPageResolution();
      // PageResolution is reported different by several Java VMs
      // To my experience the following rule meets almost any VM
      // to get the resolution that is basis for correct scaling
      if ((res > screenRes) || (res == 0))
        res = 72;
      if (printArea == null) { // use additional printer dialog
        // set orientation according to printer settings
        props.put("ORIENTATION", new Boolean(d.height > d.width));
        try {
          PrintDialog pd = new PrintDialog(dummyFrame, DemoLocal.s_Printing,
            imgSrc.width, imgSrc.height, imgSrc.horizDpi, res, img, props);
          Boolean us = (Boolean)props.get("USERSIZE");
          if ((imgSrc.printArea != null) && (us != null) && us.booleanValue())
            pd.setPrintArea(imgSrc.printArea);
          pd.show(); // show modal dialog to determine print area
          printArea = pd.getPrintArea(); // null if cancelled
        }
        catch (NoClassDefFoundError e) {
          showOkBox(DemoLocal.s_Printing, DemoLocal.s_eventModel + DemoLocal.s_notSupportedByVM + "\r\n" + DemoLocal.s_java11required);
          return true;
        }
      }
      if (printJob != null) {
        if (printArea != null) {
          imgSrc.printArea = printArea; // backup print area settings
          Graphics pg;
          pg = printJob.getGraphics();
          if (pg != null) {
            imgSrc.useProdGauge = true; // show gauge during image production
            pg.drawImage(img, printArea.x, printArea.y, printArea.width, printArea.height, parent);
            imgSrc.useProdGauge = false;
            if ((imgSrc2 != null) && (printArea2 != null)) {
              imgSrc2.useProdGauge = true; // show gauge during image production
              Image img2 = parent.createImage(imgSrc2);
              pg.drawImage(img2, printArea2.x, printArea2.y, printArea2.width, printArea2.height, parent);
              imgSrc2.useProdGauge = false;
            }
            pg.dispose(); // flush page
          }
          if (printJob == null)
            showOkBox(DemoLocal.s_security, DemoLocal.s_appletNoRightFor + DemoLocal.s_Printing);
        }
        printJob.end();
      }
    }
    return true;
  }
  */ // end of demonstration code

  boolean controlAnnotationButton(IconButton btn, int mode) {
    annotationButtonGroup.setSelected(btn, true);
    setGraphEdMode(mode);
    if (mode == GraphicEditor.POLYGONMODE)
      toolHelpLbl.setText(DemoLocal.s_PolygonHelp);
    else
      toolHelpLbl.setText("");
    if (mode == GraphicEditor.POINTERMODE) {
      setCanvasCursor(Cursor.DEFAULT_CURSOR);
      toolHelpLbl.setText(DemoLocal.s_PointerHelp);
    }
    else
      setCanvasCursor(Cursor.CROSSHAIR_CURSOR);
    return true;
  }

  // Event handling section.
  // To be compatible with Java 1.0x and Netscape Browsers prior 4.06
  // not supporting the new 1.1 event model, we use the standard
  // way overriding the handle_event method.
  // To keep the code clear we branch off to several methods
  // handling events of particular types.

  // Handle ACTION_EVENT
  // Here we find especially events generated by buttons
  //--------------------------------------------------------------------------
  boolean handleActionEvent(Event evt)
  {
    if (evt.target == btnSave) { downloadImage(); return true; }
    else if (evt.target == btnZoomIn) { zoomIn(); return true; }
    else if (evt.target == btnZoomOut) { zoomOut(); return true; }
    else if (evt.target == btnNavigation) { return navigation(); }
    else if (evt.target == btn1to1) { zoom1To1(); return true; }
    else if (evt.target == btnFitHoriz) { fitHoriz(); return true; }
    else if (evt.target == btnFitInWindow) { fitInWindow(); return true; }
    else if (evt.target == btnScaleSmooth) { toggleSmoothScaling(); return true; }
    else if (evt.target == btnAnnotations) { toggleAnnotations(); return true; }
    else if (evt.target == btnNextPage) return setNextPage();
    else if (evt.target == btnThumbnails) { toggleThumbnails(); return true; }
    else if (evt.target == btnPrevPage) return setPreviousPage();
    else if (evt.target == btnRot090) return rotate090();
    else if (evt.target == btnRot180) return rotate180();
    else if (evt.target == btnRot270) return rotate270();
    else if (evt.target == btnMirror) return mirror();
    else if (evt.target == btnFlip) return flip();
    else if (evt.target == btnInvert) return invert();
    else if (evt.target == btnPrint) return printImage(true);
    else if (evt.target == btnPrintView) return printCurrentView(true);
    else if (evt.target == btnHelp) { showKeyHelp(); return true; }
    else if (evt.target == tfPageNum) {
      try {
        return setPageNum(tbc.getPageNum());
      } catch (Exception e) { updatePageNum(); return true; }
    }
    else if (evt.target ==  tfZoom) {
      zoom(tbc.getZoomFactor(xZoom)); // use xZoom as default, if TextField contains invalid String
      return true;
    }
    else if (evt.target == btnPointer) {
      return controlAnnotationButton(btnPointer, GraphicEditor.POINTERMODE);
    }
    else if (evt.target == btnRectangle) {
      return controlAnnotationButton(btnRectangle, GraphicEditor.RECTANGLEMODE);
    }
    else if (evt.target == btnOval) {
      return controlAnnotationButton(btnOval, GraphicEditor.OVALMODE);
    }
    else if (evt.target == btnPolygon) {
      return controlAnnotationButton(btnPolygon, GraphicEditor.POLYGONMODE);
    }
    else if (evt.target == btnMemo) {
      return controlAnnotationButton(btnMemo, GraphicEditor.MEMOMODE);
    }
    else if (evt.target == btnText) {
      return controlAnnotationButton(btnText, GraphicEditor.TEXTMODE);
    }
    else if (evt.target == btnStamp) {
      return controlAnnotationButton(btnStamp, GraphicEditor.STAMPMODE);
    }
    else if (evt.target == btnShapeProps) {
      IconButton tmpBtn = annotationButtonGroup.getSelected();
      graphEd.requestShapeProps();
      props.put("STAMPS", graphEd.getStamps());
      props.put("CURRENTSTAMP", new Integer(graphEd.currentStampIdx));
      props.put("SHAPEPROPS", graphEd.currentShapeProps);
      annotationButtonGroup.setSelected(tmpBtn, true);
      return true;
    }
    else if (evt.target == btnSaveAnnot) {
      IconButton tmpBtn = annotationButtonGroup.getSelected();
      annotationSaver.saveAnnotations();
      annotationButtonGroup.setSelected(tmpBtn, true);
      return true;
    }
    ((Component)imgCanvas).requestFocus();
    return super.handleEvent(evt);
  }

  // Handle KEY_ACTION (see also handle KEY_PRESS)
  //--------------------------------------------------------------------------
  boolean handleKeyAction(Event evt)
  {
    switch (evt.key)
    {
      case Event.RIGHT:
        if (sScrPane.isSbHenabled())
        {
          int offset = sScrPane.sbHoriz.getLineIncrement();
          if (evt.modifiers == Event.CTRL_MASK)
            offset = sScrPane.sbHoriz.getPageIncrement();
          sScrPane.sbHoriz.setValue(sScrPane.sbHoriz.getValue() + offset);
          sScrPane.saveValues(); // to avoid echos
          repaint();
        }
        return true;
      case Event.LEFT:
        if (sScrPane.isSbHenabled())
        {
          int offset = sScrPane.sbHoriz.getLineIncrement();
          if (evt.modifiers == Event.CTRL_MASK)
            offset = sScrPane.sbHoriz.getPageIncrement();
          sScrPane.sbHoriz.setValue(sScrPane.sbHoriz.getValue() - offset);
          sScrPane.saveValues(); // to avoid echos
          repaint();
        }
        return true;
      case Event.DOWN:
        if (sScrPane.isSbVenabled())
        {
          sScrPane.sbVert.setValue(sScrPane.sbVert.getValue() +
                                   sScrPane.sbVert.getLineIncrement());
          sScrPane.saveValues(); // to avoid echos
          repaint();
        }
        return true;
      case Event.UP:
        if (sScrPane.isSbVenabled())
        {
          sScrPane.sbVert.setValue(sScrPane.sbVert.getValue() -
                                   sScrPane.sbVert.getLineIncrement());
          sScrPane.saveValues(); // to avoid echos
          repaint();
        }
        return true;
      case Event.PGDN:
        if (sScrPane.isSbVenabled())
        {
          if (evt.modifiers == Event.CTRL_MASK)
            sScrPane.sbVert.setValue(sScrPane.sbVert.getMaximum());
          else
            sScrPane.sbVert.setValue(sScrPane.sbVert.getValue() +
                                   sScrPane.sbVert.getPageIncrement());
          sScrPane.saveValues(); // to avoid echos
          repaint();
        }
        return true;
      case Event.PGUP:
        if (sScrPane.isSbVenabled())
        {
          if (evt.modifiers == Event.CTRL_MASK)
            sScrPane.sbVert.setValue(sScrPane.sbVert.getMinimum());
          else
            sScrPane.sbVert.setValue(sScrPane.sbVert.getValue() -
                                     sScrPane.sbVert.getPageIncrement());
          sScrPane.saveValues(); // to avoid echos
          repaint();
        }
        return true;
      case Event.HOME:
        if (sScrPane.isSbHenabled())
        {
          sScrPane.sbHoriz.setValue(sScrPane.sbHoriz.getMinimum());
          sScrPane.saveValues(); // to avoid echos
          repaint();
        }
        return true;
      case Event.END:
        if (sScrPane.isSbHenabled())
        {
          sScrPane.sbHoriz.setValue(sScrPane.sbHoriz.getMaximum());
          sScrPane.saveValues(); // to avoid echos
          repaint();
        }
        return true;
      case Event.F1:
        showKeyHelp();
        return true;
      case Event.DELETE:
        if (annotationMode() && (imgInfo.shapeLists[imgInfo.page-1].taggedCount() > 0)) {
          imgInfo.shapeLists[imgInfo.page-1].removeTaggedShapes();
          forceImageUpdate();
        }
        return true;
      default:
        return super.handleEvent(evt);
    }
  }

  // Handle KEY_PRESS (see also handle KEY_ACTION)
  //--------------------------------------------------------------------------
  boolean handleKeyPress(Event evt)
  {
    char key = Character.toUpperCase((char)evt.key);
    switch (key)
    {
      case '+':
        {  zoomIn();  return true;  }
      case '-':
        {  zoomOut();  return true;  }
      case '0': {
        if (!fitInWindow) {
          fitInWindow = true;
          fitHoriz = false;
          repaint();
        }
        return true;
      }
      case '1':
        {  zoom1To1();  return true;  }
      case '#':
        {
            toggleSmoothScaling();
          return true;
        }
      case 'I': {
        imgSrc.invert();
        forceImageUpdate(); // force image update although bound of visible area did not change
        return true;
      }
      case 'P': {
        if (cPrint.isEnabled())
          return printImage(true);
      }
      case '<': return setPreviousPage();
      case '>': return setNextPage();
      case Event.ENTER: // This event was triggered by a doubbleclick in the catalog panel
        { getSelected(); return true; }
      default:
        return handleKeyAction(evt); // MS SDK 2.0beta dispatches cursor keys with id KEY_PRESS
    }
  }

  // Handle scrollbar event
  //--------------------------------------------------------------------------
  boolean handleScrollbarEvent(Event evt) {
    boolean
      PAGESTEP = true,
      LINESTEP = false;

    ((Component)imgCanvas).requestFocus(); // return focus to imgCanvas
    switch (evt.id) {
      case Event.SCROLL_ABSOLUTE:
        sScrPane.saveValues();
        repaint();
        return true;
      case Event.SCROLL_LINE_DOWN:
        // determine if we need an adjustment of step size
        // (depends on supplier and version of Java VM)
        if (evt.target == sScrPane.sbHoriz)
          sScrPane.needStepAdjust =
            (sScrPane.sbHoriz.getValue()- sScrPane.oldHvalue
             < sScrPane.sbHoriz.getLineIncrement());
        else
          sScrPane.needStepAdjust =
            (sScrPane.sbVert.getValue() - sScrPane.oldVvalue
             < sScrPane.sbVert.getLineIncrement());
        doScrollStep(evt.target, 1, LINESTEP);
        return true;
      case Event.SCROLL_LINE_UP:
        // determine if we need an adjustment of step size
        // (depends on supplier and version of Java VM)
        if (evt.target == sScrPane.sbHoriz)
          sScrPane.needStepAdjust =
            (sScrPane.oldHvalue - sScrPane.sbHoriz.getValue()
             < sScrPane.sbHoriz.getLineIncrement());
        else
          sScrPane.needStepAdjust =
            (sScrPane.oldVvalue - sScrPane.sbVert.getValue()
             < sScrPane.sbVert.getLineIncrement());
        doScrollStep(evt.target, -1, LINESTEP);
        return true;
      case Event.SCROLL_PAGE_DOWN:
        // determine if we need an adjustment of step size
        // (depends on supplier and version of Java VM)
        if (evt.target == sScrPane.sbHoriz)
          sScrPane.needStepAdjust =
            (sScrPane.sbHoriz.getValue() - sScrPane.oldHvalue
             < sScrPane.sbHoriz.getPageIncrement());
        else
          sScrPane.needStepAdjust =
            (sScrPane.sbVert.getValue() - sScrPane.oldVvalue
             < sScrPane.sbVert.getPageIncrement());
        doScrollStep(evt.target, 1, PAGESTEP);
        return true;
      case Event.SCROLL_PAGE_UP:
        // determine if we need an adjustment of step size
        // (depends on supplier and version of Java VM)
        if (evt.target == sScrPane.sbHoriz)
          sScrPane.needStepAdjust =
            (sScrPane.oldHvalue - sScrPane.sbHoriz.getValue()
             < sScrPane.sbHoriz.getPageIncrement());
        else
          sScrPane.needStepAdjust =
            (sScrPane.oldVvalue - sScrPane.sbVert.getValue()
             < sScrPane.sbVert.getPageIncrement());
        doScrollStep(evt.target, -1, PAGESTEP);
        return true;
      default:
        return super.handleEvent(evt);
    }
  }

  // Handle mouse event
  //--------------------------------------------------------------------------
  boolean handleMouseEvent(Event evt)
  {
    if (((Component)imgCanvas instanceof ImgCanvas) && (evt.target != imgCanvas)) {
      return super.handleEvent(evt);
    }

    // translate absolute mouse coordinates relative to imgage panel
    Rectangle r = imgPanel.bounds();
    evt.translate(-r.x, -r.y);

    switch (evt.id)  {
    case Event.MOUSE_DRAG:
      if (mouseDown) {
        Graphics g = ((Component)imgCanvas).getGraphics();
        if (annotationMode() && graphEd.handleMouseDrag(mouseMove.x, mouseMove.y, evt.x, evt.y, evt.modifiers,g))
          mouseMove.move(evt.x, evt.y);
        else
          if ((evt.modifiers & Event.META_MASK) != 0) { //  panning with right mouse button
            Point p = imageOffset();
            scrollAbs((int)((p.x + mouseMove.x - evt.x) / xZoom), (int)((p.y + mouseMove.y - evt.y) / yZoom));
            mouseMove.move(evt.x, evt.y);
          } else {
            drawMouseRect(g);
            mouseMove.move(evt.x, evt.y);
            drawMouseRect(g);
          }
        g.dispose();
        return true;
      } else
        return super.handleEvent(evt);
    case Event.MOUSE_DOWN:
      // JDK 1.1(.1) will errorneously generate this event also on resizing
      // but with coordinates outside the frame, so check if clicked inside
      if (inside(evt.x, evt.y)) {
        mouseDown = true;
        mouseStart.move(evt.x, evt.y);
        mouseMove.move(evt.x, evt.y);
        Graphics g = ((Component)imgCanvas).getGraphics();
        if ((evt.modifiers & Event.META_MASK) != 0) { // right mouse button
          setCanvasCursor(Cursor.MOVE_CURSOR);
        } else { // left mouse button
          if (!annotationMode() || !graphEd.handleMouseDown(evt.x, evt.y, evt.modifiers, evt.clickCount, g)) {
            setCanvasCursor(Cursor.CROSSHAIR_CURSOR);
            drawMouseRect(g);
            if (!versionEqualGreater1_1) // old java version!
              // for Java 1.0x VMs and 256-Color displays the color palette
              // will be disturbed when we start drawing the XOR rectangle
              if ((getToolkit().getColorModel() instanceof IndexColorModel) &&
                  !paletteCorrected) {
                // color palette was disturbed - repaint with new palette (necessary one time)
                g.setColor(Color.black);
                g.setXORMode(Color.white);
                // img.flush();
                update(g);
                paletteCorrected = true;
                drawMouseRect(g);
              }
          }
        }
        g.dispose();
        return true;
      }
      else
        return super.handleEvent(evt);
    case Event.MOUSE_UP:
      // JDK 1.1(.1) will errorneously generate this event also on resizing
      // but with coordinates outside the frame
      if (mouseDown /* && inside(evt.x, evt.y) */)
      {
        if (annotationMode() && (imgInfo.shapeLists != null) && (imgInfo.shapeLists[imgInfo.page-1].taggedCount() > 0) &&
            !mouseMove.equals(mouseStart)) {
          forceImageUpdate();
        }
        else
          if ((evt.modifiers & Event.META_MASK) != 0) { // right mouse button
            setCanvasCursor(Cursor.CROSSHAIR_CURSOR);
          } else { // left mouse button
            Graphics g = ((Component)imgCanvas).getGraphics();
            if (annotationMode() && graphEd.handleMouseUp(evt.x, evt.y, evt.modifiers, g)) {
              forceImageUpdate();
            } else {
              drawMouseRect(g);
              mouseMove.move(evt.x, evt.y);
              mouseDown = false;
              zoomToRect(new PRectangle(mouseStart, mouseMove));
            }
            g.dispose();
          }
        return true;
      }
      else
        return super.handleEvent(evt);
    default:
        return super.handleEvent(evt);
    }
  }

  /**
   * The handleEvent() method receives all events generated within the frame
   * window. You can use this method to respond to window events. To respond
   * to events generated by menus, buttons, etc. or other controls in the
   * frame window but not managed by the applet, override the window's
   * action() method.
   * @param evt Event to be handled
   * @return true if event was handled
   */
  public boolean handleEvent(Event evt)
  {
    // As a modal dialog window opened from a browser applet does not
    // behave as expected, swallow event and bring dialog to front
    if (((evt.id == Event.MOUSE_DOWN) || (evt.id == Event.ACTION_EVENT) ||
         (evt.id == Event.KEY_ACTION) || (evt.id == Event.KEY_PRESS)) &&
         (ach.OkBox.currentOkBox != null)) {
      ach.OkBox.showCurrentOkBox(); return super.handleEvent(evt);
    } else
    switch (evt.id)
    {
      case Event.MOUSE_DRAG:
        return handleMouseEvent(evt);

      case Event.MOUSE_DOWN:
        return handleMouseEvent(evt);

      case Event.MOUSE_UP:
        return handleMouseEvent(evt);

      case Event.WINDOW_DESTROY:
        cleanUp();
        return true;

      case Event.ACTION_EVENT:
        return handleActionEvent(evt);

      case Event.KEY_ACTION:
        if ((evt.target == this) || (evt.target == imgCanvas))
          return handleKeyAction(evt);
        else
          return super.handleEvent(evt); // otherwise register dialog misses keys in Symantecs VM
      case Event.KEY_PRESS:
        if ((evt.target == this) || (evt.target == imgCanvas))
          return handleKeyPress(evt);
        else
          return super.handleEvent(evt); // otherwise register dialog misses keys in Symantecs VM
      case CATALOGFINISHED:
        slides = (String[])evt.arg;
        tbc.setNumPagesText(slides.length);
        return true;

      default:
        if ((evt.target instanceof Scrollbar) || (evt.target instanceof ScrollComponent))
          handleScrollbarEvent(evt);
        else {
          return super.handleEvent(evt);
        }
    }
    // we should never get here...
    return super.handleEvent(evt);
  }

  /**
   * The action handler should not be necessary
   * as all events should be handled already by handleEvent()
   * @param evt Event to be handled
   * @param arg the argument object
   * @return true if event was handled
   */
  public boolean action(Event evt, Object arg)
  {
    System.err.println("Unexpected action: " + evt.toString());
    return super.action(evt, arg);
  }


  // method to track image loading process
  //--------------------------------------------------------------------------
  void trackImage(Image img)
  {
    final int id = 42;
    // include MediaTracker to make sure image is loaded before continuing
    MediaTracker tracker = new MediaTracker(this);
    tracker.addImage(img,id);
    try { tracker.waitForID(id); }
    catch(InterruptedException ex) {  }
  }

  // method to track image loading process of EMemImageSources
  //--------------------------------------------------------------------------
  void trackEMemImageSource(EMemImageSource imgSrc)
  {
    while (!imgSrc.isInitialized() && (imgSrc.errorMsg == null)) {
      try { Thread.sleep(50);
      } catch (InterruptedException ie) { imgSrc.abort(); }
    }
    while (!imgSrc.isComplete() && (imgSrc.errorMsg == null)) {
      try { Thread.sleep(100);
      } catch (InterruptedException ie) { imgSrc.abort(); }
    }
  }

  /**
   * The createImageProducer() method of class ImageInfo works
   * asynchronously and will call the imageProducerUpdate method
   * of an ImgProdObserver (implemented by this viewer class!).
   * Finally showPanel() will be called by this method
   * @param imgSrc ImageProducer that was created asynchronously
   * @param errorMsg String that contains an error message if an error occured during ImageProducer creation
   * @param imgInfo the ImageInfo object containing the imgName string (used if error is displayed)
   */
  public void imageProducerUpdate(ImageProducer imgSrc, String errorMsg, ImageInfo imgInfo) {
    boolean useGauge = true; // use gauge

    if ((errorMsg != null) && (errorMsg.length() > 0)) {
      // showOkBox(DemoLocal.s_error, errorMsg);
      this.errorMsg = DemoLocal.s_fileNotFoundOrError + "\r\n" + errorMsg;
      this.imgSrc = null;
    }
    else try {
      if (imgSrc instanceof EMemImageSource)
        this.imgSrc = (EMemImageSource)imgSrc;
      else { // create an EMemImageSource from ImageProducer
        // and wait until image is loaded completely
        EMemImageSource tmpImgSrc =  new EMemImageSource(imgSrc, useGauge);
        trackEMemImageSource(tmpImgSrc);
        if (tmpImgSrc.errorMsg != null) {
          errorMsg = tmpImgSrc.errorMsg;
          showOkBox(DemoLocal.s_error, errorMsg + ": " + imgInfo.imgName);
          tmpImgSrc = null;
        } else
        if (tmpImgSrc.height <= 0) {
          showOkBox(DemoLocal.s_error, DemoLocal.s_fileNotFoundOrError + " " + imgInfo.imgName);
          tmpImgSrc = null;
        }
        this.imgSrc = tmpImgSrc;
      }
    } catch (Throwable e) {
        showOkBox(DemoLocal.s_error, DemoLocal.s_fileNotFoundOrError + " " + imgInfo.imgName);
    }
    if (this.imgSrc != null) {
      if (doInvert)
        this.imgSrc.invert();
      if (doRotate != 0) {
        switch (doRotate) {
          case 1:
            this.imgSrc = this.imgSrc.getRotated090();
            applyShapeListRotation(1);
            break;
          case 2:
            this.imgSrc.rotate180();
            applyShapeListRotation(2);
            break;
          case 3:
            this.imgSrc = this.imgSrc.getRotated270();
            applyShapeListRotation(3);
            break;
        }
      }
      if (unskewAllowed)
        if (this.imgSrc.horizDpi != this.imgSrc.vertDpi)
          this.imgSrc = this.imgSrc.unskew();
      this.imgInfo.page = this.imgSrc.page;
      this.imgInfo.numPages = this.imgSrc.numPages;
    } else {
      this.imgInfo.page = 1;
      this.imgInfo.numPages = 0;
    }
    if ((this.imgInfo.shapeLists != null) &&
        (this.imgInfo.shapeLists.length >= imgInfo.page) &&
        (this.imgInfo.shapeLists[this.imgInfo.page-1] != null)) {// set state of annotation button if annotations exist
      enableAnnotations(imgInfo.shapeLists[imgInfo.page-1]);
    } else {
      disableAnnotations();
    }
    showPanel();
    if (printJob != null) {
      try {
        // due to precompilation of some code that will never be executed in MS IE...
        autoPrintImage(printJob); // ...currently causes an IOException in MS IE (trying to read PrinterJob)
      } catch (Throwable e) { }
      printPageNum = nextPageFromPrintPageSelection(printPageSelection, printPageNum);
      if (printPageNum >= 0)
        setPageNum(printPageNum);
      else { // printPageNum < 0 signals end of job
        if (printJob instanceof PrintJob)
          ((PrintJob)printJob).end();
        else {
          PrintImage2.printBook(printJob, book);
          book = null;
        }
        printJob = null;
      }
    }
    if (cmdLinePrint) {
      printImage(false); // start synchronous printout
      cmdLinePrint = false;
      cleanUp();
    }
  }

  void checkAnnotationsModified() {
    boolean annotationsModified = false;
    if (imgInfo.shapeLists != null)
      for (int i=0; i<imgInfo.shapeLists.length; i++)
        if (imgInfo.shapeLists[i].isModified())
          annotationsModified = true;
    if (annotationsModified) {
      YesNoBox yn = new YesNoBox(dummyFrame,DemoLocal.s_Annotations, DemoLocal.s_Save + "?");
      yn.show();
      if (yn.getAnswer() == YesNoBox.YES)
        annotationSaver.saveAnnotations();
      yn.dispose();
    }
  }

  /**
   * Comfortable method to load image (using format autodection);
   * runs createImageProducer() and asynchronously calls back imageProducerUpdate()
   * which finally calls showPanel() to display the image.
   * @param imgInfo an instance of class ImageInfo used to specify URL or filename
   * of the image and set the desired pagenumber (if a multipage image is loaded).
   */
  public void loadImage(ImageInfo imgInfo)
  {
    setCanvasCursor(Cursor.WAIT_CURSOR);
    checkAnnotationsModified();
    errorMsg =  null;
    Toolkit toolkit = getToolkit();
    this.imgInfo = imgInfo;
    boolean mode = false; // do not load asynchronously
    // Please look at the source code for class ImageInfo to learn
    // how to detect and load several image types!
    System.gc();
    imgInfo.setAsynchronousTools(toolkit, this, mode);
    new Thread(imgInfo).start(); // runs createImageProducer() and calls back imageProducerUpdate()
    // if lightweight components are used, it might be necessary to use
    // the invokeLater-method instead of Thread(...).start()
    // but this is no general rule...
    // javax.swing.SwingUtilities.invokeLater(new Thread(imgInfo));
    if (navDialog != null) {
      navDialog.dispose();
      navDialog = null;
    }
  }

  /**
   * Callback function to monitor image drawing process
   * will be called several times during drawImage (why not called in JDK 1.1?)
   * @see java.awt.Component#imageUpdate(java.awt.Image, int, int, int, int, int)
   */
  public boolean imageUpdate(Image img, int flags, int x, int y, int w, int h)
  {
    if ( (flags & ABORT) != 0 ) // Image fetcher interrupted - try repaint
      {  repaint(); return false;  }
    return super.imageUpdate(img, flags, x, y, w, h);
  }

  // Draw mouse rectangle
  //--------------------------------------------------------------------------
  void drawMouseRect(Graphics g)
  {
    g.setColor(Color.black);
    g.setXORMode(Color.white);
    PRectangle rect = new PRectangle(mouseStart, mouseMove);
    g.drawRect(rect.x, rect.y, rect.width, rect.height);
  }

  /**
   * Paint Handler of this component
   * @param g Graphics context to be painted
   */
  public void paint(Graphics g)
  {
    if (!((Component)imgCanvas instanceof ImgComponent)) // heavyweight component needs super.paint first
      super.paint(g);
    try {
    if ((imgSrc != null) && (imgCanvas != null) && (imgSrc.height > 0))
    {
      frameUpdated = true;
      if (bufferedAnnotations != null) {
        imgInfo.shapeLists = new Annotations().createShapeLists(bufferedAnnotations);
        bufferedAnnotations = null;
        enableAnnotations(imgInfo.shapeLists[imgInfo.page-1]);
        showPanel(); // revalidates panel components to display annotation editor toolbar
      }
      if (showCatalog && (catSlides != null) && (catSlides.length > 0)) {
          catPanel.addImageList(catSlides, new Dimension(thumbNailWidth,thumbNailHeight), Image.SCALE_SMOOTH);
          catSlides = null;
      }
      int imgWidth = imgSrc.width;
      int imgHeight = imgSrc.height;
      if (fitHoriz) // scales horizontally to window width
      {
        // sScrPane.disableScrollPane(); // removed because it blocks absolute scrolling in horiz-mode
        if ((Component)imgCanvas instanceof ImgComponent)
          validate(); // make sure imgCanvas uses space of invisible scrollbars
        int visiWidth = sScrPane.getVisibleWidth(imgPanel);
        int visiHeight = sScrPane.getVisibleHeight(imgPanel);
        // if (visiHeight < imgHeight)
        //   visiWidth -= sScrPane.sbVert.minimumSize().width; // correction only required when scrollbars previously disabled
        xZoom = (double)visiWidth / imgWidth;
        yZoom = xZoom;
      } else
      if (fitInWindow) // scales proportionally to fit into window
      {
        sScrPane.disableScrollPane();
        if ((Component)imgCanvas instanceof ImgComponent)
          validate(); // make sure imgCanvas uses space of invisible scrollbars
        xZoom = (double)sScrPane.getVisibleWidth(imgPanel) / imgWidth;
        if (xZoom * imgHeight <= (double)sScrPane.getVisibleHeight(imgPanel))
          yZoom = xZoom;
        else {
          yZoom = (double)sScrPane.getVisibleHeight(imgPanel) / imgHeight;
          xZoom = yZoom;
        }
      }
      int zoomedWidth = (int)(xZoom * imgWidth);
      int zoomedHeight = (int)(yZoom * imgHeight);
      // Due to an error in Microsofts (Java VM 1.0) scrollbar-handling values must be restored again!
      sScrPane.restoreValues();
      sScrPane.updateScrollbars(imgPanel, new Dimension(zoomedWidth, zoomedHeight));
      int x = - sScrPane.getHorizValue(imgPanel, zoomedWidth);
      int y = - sScrPane.getVertValue(imgPanel, zoomedHeight);
      if ((imgInfo.shapeLists != null) && (imgInfo.shapeLists.length >= imgInfo.page) && (imgInfo.shapeLists[imgInfo.page-1] != null))
        imgInfo.shapeLists[imgInfo.page-1].applyBounds(new Rectangle(x, y, zoomedWidth, zoomedHeight));
      int visiWidth = sScrPane.getVisibleWidth(imgPanel);
      int visiHeight = sScrPane.getVisibleHeight(imgPanel);
      if ((((Component)imgCanvas).size().width != visiWidth) || (((Component)imgCanvas).size().height != visiHeight))
        imgCanvas.resize(visiWidth, visiHeight);
      imgCanvas.checkFocus(); // only needed in java version < 1.1 in order to get keyboard events
      newTmpImgRect = getRealCoordinates(
        new Rectangle(0, 0, visiWidth, visiHeight));
      newTmpImgRect = newTmpImgRect.intersection(new Rectangle(0,0,imgWidth,imgHeight));

      boolean useTmpImg = false;

      // if visible area is smaller than img size
      // then create cropped temporary image
      if (((xZoom >= 1) || (yZoom >= 1))
          && ((newTmpImgRect.width < imgWidth) ||
            (newTmpImgRect.height < imgHeight))) {
        useTmpImg = true;
        if (!newTmpImgRect.equals(tmpImgRect)) {
          // old tmpImg can not be reused
          tmpImgRect = newTmpImgRect;
          int oldCursor = getCanvasCursor();
          setCanvasCursor(Cursor.WAIT_CURSOR);
          try {
            // if (tmpImg != null)
            //  tmpImg.flush();

            // in old Java VMs (< 1.1), scaling by java.awt is faster
            if (( !versionEqualGreater1_1 ) || ((xZoom == 1) && (yZoom == 1))) {
              // Java 1.0.2 has no method getScaledInstance()
              // tmpImage will be scaled later by drawImage...
              tmpImg = createImage(imgSrc.extractMemImgRegion(tmpImgRect.x, tmpImgRect.y,
                tmpImgRect.width, tmpImgRect.height));
            }
            else
              tmpImg = createImage(imgSrc.extractMemImgRegion(tmpImgRect.x, tmpImgRect.y,
                tmpImgRect.width, tmpImgRect.height, visiWidth, visiHeight));
          } finally {
            setCanvasCursor(oldCursor);
          }

        }
        if (newTmpImgRect.width < imgWidth) x = 0;
        else visiWidth = zoomedWidth;
        if (newTmpImgRect.height < imgHeight) y = 0;
        else visiHeight = zoomedHeight;
      }
      // if image is shrinked, scale before dithering (with java >= 1.1)
      // Java < 1.1 scales before dithering and is faster for small images
      // but consumes too much memory for huge images (if not cropped)
      if ((xZoom < 1) || (yZoom < 1)) {
        useTmpImg = true;
        if (!newTmpImgRect.equals(tmpImgRect) ||
          ((imgSrc.calcPixelSize() == 1) && (tmpImgRect.width == imgSrc.width) &&
           (tmpImg.getWidth(this) != zoomedWidth))) {
          // old tmpImg can not be reused
          // also if we have a bilevel image and zoomFactor was changed, force rescale
          tmpImgRect = newTmpImgRect;
          // if (tmpImg != null)
          //  tmpImg.flush();
          int oldCursor = getCanvasCursor();
          setCanvasCursor(Cursor.WAIT_CURSOR);
          try {
            EMemImageSource tmpImgSrc;
            if ((newTmpImgRect.width < imgWidth) && (newTmpImgRect.height < imgHeight)) {
              // we dont need the whole image
              // use getScaledInstance only if not bilevel image
              if ((scaledInstanceSupported) && (imgSrc.calcPixelSize() > 1)) {
                try {
                  tmpImgSrc = new EMemImageSource(createImage(imgSrc.extractMemImgRegion(tmpImgRect.x, tmpImgRect.y,
                    tmpImgRect.width, tmpImgRect.height)).getScaledInstance(
                      visiWidth, visiHeight, scaleHint).getSource(), false);
                  trackEMemImageSource(tmpImgSrc);
                  tmpImg = createImage(tmpImgSrc);
                } catch (NoSuchMethodError e)
                {  // getScaledInstance not supported by several Symantec Bytecode Engines (Netscape Navigator)
                  // use same logic as below (else)
                  tmpImg = createImage(imgSrc.extractMemImgRegion(tmpImgRect.x, tmpImgRect.y,
                    tmpImgRect.width, tmpImgRect.height, visiWidth, visiHeight));
                  scaledInstanceSupported = false;
                  showOkBox("Antialiasing", DemoLocal.s_scaleSmooth + DemoLocal.s_notSupportedByVM);
                }
              }
              else
                tmpImg = createImage(imgSrc.extractMemImgRegion(tmpImgRect.x, tmpImgRect.y,
                  tmpImgRect.width, tmpImgRect.height, visiWidth, visiHeight));
            }
            else // getScaledInstance is faster if we need the whole image
              if ((scaledInstanceSupported) && (imgSrc.calcPixelSize() > 1)) {
                try {
                  tmpImgSrc = new EMemImageSource(createImage(imgSrc).getScaledInstance(zoomedWidth, zoomedHeight, scaleHint).getSource(), false);
                  trackEMemImageSource(tmpImgSrc);
                  tmpImg = createImage(tmpImgSrc);
                } catch (NoSuchMethodError e)
                { // getScaledInstance not supported by several Symantec Bytecode Engines (Netscape Navigator)
                  // use same logic as below (else)
                  tmpImg = createImage(imgSrc.extractMemImgRegion(0, 0,
                    imgWidth, imgHeight, zoomedWidth, zoomedHeight));
                  scaledInstanceSupported = false;
                  showOkBox("Antialiasing", DemoLocal.s_scaleSmooth + DemoLocal.s_notSupportedByVM);
                }
              }
              else
                tmpImg = createImage(imgSrc.extractMemImgRegion(0, 0,
                  imgWidth, imgHeight, zoomedWidth, zoomedHeight));
          } finally {
            setCanvasCursor(oldCursor);
          }
        }
        if ((newTmpImgRect.width < imgWidth) &&
            (newTmpImgRect.height < imgHeight))
        {  x = 0;  y = 0;  }
        else
        {  visiWidth = zoomedWidth;  visiHeight = zoomedHeight;  }
      }

      // calculate area that is used by image (to be excluded from background)
      // should images with transparent color be painted without background pattern?
      // if ((imgSrc.colorModel instanceof IndexColorModel) &&
      //    (((IndexColorModel)imgSrc.colorModel).getTransparentPixel() >= 0))
      //  imgCanvas.setBackgroundExclusion(0, 0, 0, 0);
      // else
      {
        imgCanvas.setBackgroundExclusion(x, y, imgWidth, imgHeight);
        if ((xZoom != 1) || (yZoom != 1))
          if (useTmpImg)
            imgCanvas.setBackgroundExclusion(x, y, visiWidth, visiHeight);
          else
            imgCanvas.setBackgroundExclusion(x, y, zoomedWidth, zoomedHeight);
      }


      if ((xZoom == 1) && (yZoom == 1)) // no scaling
        if (useTmpImg)
          imgCanvas.drawImage(tmpImg, x, y);
        else
          imgCanvas.drawImage(createImage(imgSrc), x, y);
      else
        if (useTmpImg)
        {
          if ((tmpImg.getWidth(this) == visiWidth) && (tmpImg.getHeight(this) == visiHeight))
            imgCanvas.drawImage(tmpImg, x, y);
           else // tmpImg needs to be scaled
            imgCanvas.drawImage(tmpImg, x, y, visiWidth, visiHeight, scaleHint);
        }
        else
          imgCanvas.drawImage(createImage(imgSrc), x, y, zoomedWidth, zoomedHeight, scaleHint);
      if (useTmpImg)
        System.gc();

      if (annotationMode() && (imgInfo.shapeLists.length >= imgInfo.page)) {
        imgCanvas.drawShapes(imgInfo.shapeLists[imgInfo.page-1]);
      }
      else
        imgCanvas.drawShapes(null);

      // control visibility of buttons
      if ((imgInfo.numPages > imgInfo.page) ||
        ((slides != null) && (slideIdx < slides.length-1)))
        if (!cNextPage.isEnabled())
          if (versionEqualGreater1_1) cNextPage.setEnabled(true); else cNextPage.enable();
        else ;
      else
        if (cNextPage.isEnabled())
          if (versionEqualGreater1_1) cNextPage.setEnabled(false); else cNextPage.disable();
      if ((imgInfo.page > 1) || ((slides != null) && (slideIdx > 0)))
        if (!cPrevPage.isEnabled())
          if (versionEqualGreater1_1) cPrevPage.setEnabled(true); else cPrevPage.enable();
        else ;
      else
        if (cPrevPage.isEnabled())
          if (versionEqualGreater1_1) cPrevPage.setEnabled(false); else cPrevPage.disable();
      tbc.setZoomText(xZoom);
      if (scaleHint == Image.SCALE_SMOOTH){
        btnScaleSmooth.setToggleState(true);
      }
      if (isAncestorOf(catPanel) && showCatalog){
        cThumbnails.enable();
        btnThumbnails.setToggleState(true);
      }
      else
        btnThumbnails.setToggleState(false);

      if (navDialog != null) {
        if (navDialog.isShowing())
          navDialog.setViewRect(tmpImgRect);
        else
          navDialog = null;
      }
    } // end if
    } // end try
    catch (NullPointerException ne) { }

    // display error message if available
    if (errorMsg != null)
    {
      imgCanvas.setBackgroundExclusion(0, 0, size().width, size().height);
      g = ((Component)imgCanvas).getGraphics();
      g.clearRect(0, 0, size().width, size().height);
      imgCanvas.drawError(errorMsg);
      System.err.println(errorMsg);
    }
    if ((Component)imgCanvas instanceof ImgComponent) // if not heavyweight
      super.paint(g);                      // update embedded components now
  }

  /**
   * Override component's update method to suppress clearing of the screen
   * @param g Graphics context to be painted
   */
  public void update(Graphics g)
  {
    try { paint(g);  }
    catch (NullPointerException npe) { }
  }

}
